# System Imports
import os
import re
import inspect
import getpass
import commands
from GangaCore.Utility.ColourText import ANSIMarkup, overview_colours

# Global Variable to enable Job Sharing mechanism required in GANGA SWAN INTEGRATION.
# If environment variable GANGA_SWAN_INTEGRATION is present enable this mechanism.
GANGA_SWAN_INTEGRATION = "GANGA_SWAN_INTEGRATION" in os.environ

# Global Functions
def getLCGRootPath():

    lcg_release_areas = {'afs' : '/afs/cern.ch/sw/lcg/releases/LCG_79',
    'cvmfs' : '/cvmfs/lhcb.cern.ch/lib/lcg/releases/LCG_79'}

    ## CAUTION This could be sensitive to mixed AFS/CVMFS running but I doubt this setup is common or likely
    myCurrentPath = os.path.abspath(inspect.getfile(inspect.currentframe()))

    if myCurrentPath[:4].upper() == '/AFS':
        return lcg_release_areas['afs']
    elif myCurrentPath[:6].upper() == '/CVMFS':
        return lcg_release_areas['cvmfs']
    else:
        return ''

# ------------------------------------------------
# store Ganga version based on new git tag for this file
_gangaVersion = '7.1.13'
_development = False

# store a path to Ganga libraries
_gangaPythonPath = os.path.dirname(os.path.dirname(__file__))


# grab the hostname
try:
    from GangaCore.Utility.util import hostname
    hostname = hostname()
except Exception as x:  # fixme: use OSError instead?
    hostname = 'localhost'


# ------------------------------------------------
# Setup all configs for this module
from GangaCore.Utility.Config import makeConfig, getConfig, expandvars


# ------------------------------------------------
# Logging
log_config = makeConfig("Logging", """control the messages printed by Ganga
The settings are applied hierarchically to the loggers. Ganga is the name of the top-level logger which
applies by default to all GangaCore.* packages unless overriden in sub-packages.
You may define new loggers in this section.
The log level may be one of: CRITICAL ERROR WARNING INFO DEBUG
""", is_open=True)

# FIXME: Ganga WARNING should be turned into INFO level when the messages
# are reviewed in all the code
log_config.addOption('Ganga', "INFO", "top-level logger")
log_config.addOption('GangaCore.Runtime.bootstrap', "INFO", 'FIXME')
log_config.addOption('GangaCore.GPIDev', "INFO", "logger of GangaCore.GPIDev.* packages")
log_config.addOption('GangaCore.Utility.logging', "WARNING", "logger of the Ganga logging package itself (use with care!)")
log_config.addOption('_format', "NORMAL", "format of logging messages: TERSE,NORMAL,VERBOSE,DEBUG")
log_config.addOption('_colour', True, "enable ASCII colour formatting of messages e.g. errors in red")
log_config.addOption('_logfile', "~/.ganga.log", "location of the logfile")
log_config.addOption('_logfile_size', 100000,
                 "the size of the logfile (in bytes), the rotating log will never exceed this file size")  # 100 K
log_config.addOption('_interactive_cache', True,
                 'if True then the cache used for interactive sessions, False disables caching')
log_config.addOption('_customFormat', "", "custom formatting string for Ganga logging\n e.g. '%(name)-35s: %(levelname)-8s %(message)s'")

# ------------------------------------------------
# System
# the system variables (such as VERSION) are put to DEFAULTS section of the config module
# so you can refer to them in the config file
# additionally they will be visible in the (write protected) [System]
# config module - write protection is done in bootstrap at present though this should be changed
import GangaCore.Utility.files
config_path = GangaCore.Utility.files.expandfilename(
    os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), '..')) )
sys_config = makeConfig('System', "parameters of this ganga session (read-only)", cfile=False)
sys_config.addOption('GANGA_VERSION', _gangaVersion, '')
sys_config.addOption('GANGA_PYTHONPATH', _gangaPythonPath, 'location of the ganga core packages')
sys_config.addOption('GANGA_CONFIG_PATH', config_path + '/','site/group specific configuration files as specified by --config-path or GANGA_CONFIG_PATH variable')
sys_config.addOption('GANGA_CONFIG_FILE', os.path.expanduser('~/.gangarc'), 'current user config file used')
sys_config.addOption('GANGA_HOSTNAME', hostname, 'local hostname where ganga is running')

# ------------------------------------------------
# Configuration
# the SCRIPTS_PATH must be initialized before the config files are loaded
# for the path to be correctly prepended
conf_config = makeConfig("Configuration", "global configuration parameters.\nthis is a catch all section.")
conf_config.addOption('SCRIPTS_PATH', 'Ganga/scripts', """the search path to scripts directory.
When running a script from the system shell (e.g. ganga script) this path is used to search for script""")

conf_config.addOption('LOAD_PATH', '', "the search path for the load() function")
conf_config.addOption('RUNTIME_PATH', '',
"""path to runtime plugin packages where custom handlers may be added.
Normally you should not worry about it.
If an element of the path is just a name (like in the example below)
then the plugins will be loaded using current python path. This means that
some packages such as GangaTest may be taken from the release area.""",
        examples="RUNTIME_PATH = GangaGUIRUNTIME_PATH = /my/SpecialExtensions:GangaTest ")

conf_config.addOption('TextShell', 'IPython', """ The type of the interactive shell: IPython (cooler) or Console (limited)""")
conf_config.addOption('StartupGPI', '', 'block of GPI commands executed at startup')
conf_config.addOption('ReleaseNotes', True, 'Flag to print out the relevent subsection of release notes for each experiment at start up')
conf_config.addOption('used_versions_path', '~/.cache/Ganga/', 'Path to the directory to store the file listing the used ganga versions')
conf_config.addOption('gangadir', expandvars(None, '~/gangadir'),
                 'Location of local job repositories and workspaces. Default is ~/gangadir but in somecases (such as LSF CNAF) this needs to be modified to point to the shared file system directory.', filter=GangaCore.Utility.Config.expandvars)
conf_config.addOption('repositorytype', 'LocalXML', 'Type of the repository.', examples='LocalXML')
conf_config.addOption('lockingStrategy', 'UNIX', 'Type of locking strategy which can be used. UNIX or FIXED . default = UNIX')
conf_config.addOption('workspacetype', 'LocalFilesystem',
                 'Type of workspace. Workspace is a place where input and output sandbox of jobs are stored. Currently the only supported type is LocalFilesystem.')
conf_config.addOption('user', getpass.getuser(),
    'User name. The same person may have different roles (user names) and still use the same gangadir. Unless explicitly set this option defaults to the real user name.')
conf_config.addOption('resubmitOnlyFailedSubjobs', True,
                 'If TRUE (default), calling job.resubmit() will only resubmit FAILED subjobs. Note that the auto_resubmit mechanism will only ever resubmit FAILED subjobs.')
conf_config.addOption('SMTPHost', 'localhost', 'The SMTP server for notification emails to be sent, default is localhost')
conf_config.addOption('deleteUnusedShareDir', 'always',
                 'If set to ask the user is presented with a prompt asking whether Shared directories not associated with a persisted Ganga object should be deleted upon Ganga exit. If set to never, shared directories will not be deleted upon exit, even if they are not associated with a persisted Ganga object. If set to always (the default), then shared directories will always be deleted if not associated with a persisted Ganga object.')

conf_config.addOption('autoGenerateJobWorkspace', False, 'Autogenerate workspace dirs for new jobs')

conf_config.addOption('NoAfsToken', False, 'Do not require an AFS token when running on an AFS filesystem. Not recommended!')

# add named template options
conf_config.addOption('namedTemplates_ext', 'tpl',
                 'The default file extension for the named template system. If a package sets up their own by calling "establishNamedTemplates" from python/Ganga/GPIDev/Lib/Job/NamedJobTemplate.py in their ini file then they can override this without needing the config option')
conf_config.addOption('namedTemplates_pickle', False,
                 'Determines if named template system stores templates in pickle file format (True) or in the Ganga streamed object format (False). By default streamed object format which is human readable is used. If a package sets up their own by calling "establishNamedTemplates" from python/Ganga/GPIDev/Lib/Job/NamedJobTemplate.py in their ini file then they can override this without needing the config option')

# add server options
conf_config.addOption('ServerPort', 434343, 'Port for the Ganga server to listen on')
conf_config.addOption('ServerTimeout', 60, 'Timeout in minutes for auto-server shutdown')
conf_config.addOption('ServerUserScript', "", "Full path to user script to call periodically. The script will be executed as if called within Ganga by 'execfile'.")
conf_config.addOption('ServerUserScriptWaitTime', 300, "Time in seconds between executions of the user script")

conf_config.addOption('confirm_exit', True, 'Ask the user on exit if we should exit, (this is passed along to IPython)')
conf_config.addOption('force_start', False, 'Ignore disk checking on startup')

conf_config.addOption('DiskIOTimeout', 45, 'Time in seconds before a ganga session (lock file) is treated as a zombie and removed')

# runtime warnings issued by the interpreter may be suppresed
conf_config.addOption('IgnoreRuntimeWarnings', False, "runtime warnings issued by the interpreter may be suppresed")

conf_config.addOption('Batch', 'LSF', 'default batch system')

conf_config.addOption('AutoStartReg', True, 'AutoStart the registries, needed to access any jobs in registry therefore needs to be True for 99.999% of use cases')
# ------------------------------------------------
# IPython
ipython_config = makeConfig('TextShell_IPython', '''IPython shell configuration''')
ipython_config.addOption('colourscheme','LightBG', 'Colour scheme to be used by iPython. Options are LightBG, Linux, Neutral, NoColor' )

# ------------------------------------------------
# Shell
makeConfig("Shell", "configuration parameters for internal Shell utility.")

# ------------------------------------------------
# Queues
queues_config = makeConfig("Queues", "configuration section for the queues")
queues_config.addOption('Timeout', None, 'default timeout for queue generated processes')
queues_config.addOption('NumWorkerThreads', 5, 'default number of worker threads in the queues system')

# ------------------------------------------------
# Plugins
default_plugins_cfg = makeConfig('Plugins', '''General control of plugin mechanism.
Set the default plugin in a given category.
For example:
default_applications = DaVinci
default_backends = LCG
''')

# ------------------------------------------------
# GPI Semantics
gpi_config = makeConfig('GPI_Semantics',
                       'Customization of GPI behaviour. These options may affect the semantics of the Ganga GPI interface (what may result in a different behaviour of scripts and commands).')
gpi_config.addOption('job_submit_keep_going', False,
                    'Keep on submitting as many subjobs as possible. Option to j.submit(), see Job class for details')
gpi_config.addOption('job_submit_keep_on_fail', False,
                    'Do not revert job to new status even if submission failed. Option to j.submit(), see Job class for details')

# ------------------------------------------------
# PollThread
poll_config = makeConfig('PollThread', 'background job status monitoring and output retrieval')
poll_config.addOption('repeat_messages', False, 'if 0 then log only once the errors for a given backend and do not repeat them anymore')
poll_config.addOption('autostart', True, 'enable monitoring automatically at startup, in script mode monitoring is disabled by default, in interactive mode it is enabled', type=type(True))  # enable monitoring on startup
poll_config.addOption('autostart_monThreads', True, 'enable populating of the monitoring worker threads')
poll_config.addOption('enable_multiThreadMon', True, 'enable multiple threads to be used for running monitoring tasks')
poll_config.addOption('base_poll_rate', 2, 'internal supervising thread', hidden=1)
poll_config.addOption('MaxNumResubmits', 5, 'Maximum number of automatic job resubmits to do before giving')
poll_config.addOption('MaxFracForResubmit', 0.25, 'Maximum fraction of failed jobs before stopping automatic resubmission')
poll_config.addOption('autoKillThreshold', 20, 'Maximum number of failed subjobs before a job is automatically killed by the monitoring.')
poll_config.addOption('update_thread_pool_size', 5, 'Size of the thread pool. Each threads monitors a specific backaend at a given time. Minimum value is one, preferably set to the number_of_backends + 1')
poll_config.addOption('default_backend_poll_rate', 30, 'Default rate for polling job status in the thread pool. This is the default value for all backends.')
poll_config.addOption('Local', 10, 'Poll rate for Local backend.')
poll_config.addOption('LCG', 30, 'Poll rate for LCG backend.')
poll_config.addOption('Condor', 30, 'Poll rate for Condor backend.')
poll_config.addOption('gLite', 30, 'Poll rate for gLite backend.')
poll_config.addOption('LSF', 20, 'Poll rate for LSF backend.')
poll_config.addOption('PBS', 20, 'Poll rate for PBS backend.')
poll_config.addOption('Dirac', 50, 'Poll rate for Dirac backend.')
poll_config.addOption('Panda', 50, 'Poll rate for Panda backend.')

# Note: the rate of this callback is actually
# MAX(base_poll_rate,callbacks_poll_rate)
poll_config.addOption('creds_poll_rate', 30, "The frequency in seconds for credentials checker")
poll_config.addOption('diskspace_poll_rate', 30, "The frequency in seconds for free disk checker")
poll_config.addOption('DiskSpaceChecker', "", "disk space checking callback. This function should return False when there is no disk space available, True otherwise")
poll_config.addOption('max_shutdown_retries', 5, 'OBSOLETE: this option has no effect anymore')
poll_config.addOption('numParallelJobs', 25, 'Number of Jobs to update the status for in parallel')

poll_config.addOption('forced_shutdown_policy', 'session_type',
                 'If there are remaining background activities at exit such as monitoring, output download Ganga will attempt to wait for the activities to complete. You may select if a user is prompted to answer if he wants to force shutdown ("interactive") or if the system waits on a timeout without questions ("timeout"). The default is "session_type" which will do interactive shutdown for CLI and timeout for scripts.')
poll_config.addOption('forced_shutdown_timeout', 60,
                 "Timeout in seconds for forced Ganga shutdown in batch mode.")
poll_config.addOption('forced_shutdown_prompt_time', 10,
                 "User will get the prompt every N seconds, as specified by this parameter.")
poll_config.addOption('forced_shutdown_first_prompt_time', 5,
                 "User will get the FIRST prompt after N seconds, as specified by this parameter. This parameter also defines the time that Ganga will wait before shutting down, if there are only non-critical threads alive, in both interactive and batch mode.")

import sys
poll_config.addOption('HeartBeatTimeOut', sys.maxint, 'Time before the user gets the warning that a thread has locked up due to failing to update the heartbeat attribute')

poll_config.addOption('autoCheckCredentials', True, 'Check credentials using the monitoring loop')

# ------------------------------------------------
# Feedback
feedback_config = makeConfig('Feedback', 'Settings for the Feedback plugin. Cannot be changed during the interactive Ganga session.')
feedback_config.addOption('uploadServer', 'http://gangamon.cern.ch/django/errorreports', 'The server to connect to')

# ------------------------------------------------
# Associations
assoc_config = makeConfig(
    "File_Associations", 'Default associations between file types and file-viewing commands. The name identifies the extension and the value the commans. New extensions can be added. A single & after the command indicates that the process will be started in the background. A && after the command indicates that a new terminal will be opened and the command executed in that terminal.', is_open=True)

assoc_config.addOption("newterm_command", "xterm",
                 'Command for opening a new terminal (xterm, gnome-terminal, ...')
assoc_config.addOption("newterm_exeopt", "-e",
                 'Option to give to a new terminal to tell it to execute a command.')
assoc_config.addOption(
    "listing_command", "ls -ltr", 'Command for listing the content of a directory')
assoc_config.addOption('fallback_command', 'less',
                 'Default command to use if there is no association with the file type')
assoc_config.addOption('htm', 'firefox &', 'Command for viewing html files.')
assoc_config.addOption('html', 'firefox &', 'Command for viewing html files.')
assoc_config.addOption('root', 'root.exe &&', 'Command for opening ROOT files.')
assoc_config.addOption('tar', 'file-roller &', 'Command for opening tar files.')
assoc_config.addOption('tgz', 'file-roller &', 'Command for opening tar files.')

# ------------------------------------------------
# Root
root_config = makeConfig('ROOT', "Options for Root backend")
## Not needed when we can't do option substitution internally but support it at the .gangarc level!!!!! 27-09-2015 rcurrie
#config.addOption('lcgpath', getLCGRootPath(), 'Path of the LCG release that the ROOT project and it\'s externals are taken from')
root_config.addOption('arch', 'x86_64-slc6-gcc48-opt', 'Architecture of ROOT')
## Auto-Interporatation doesn't appear to work when setting the default value
#config.addOption('location', '${lcgpath}/ROOT/${version}/${arch}/', 'Location of ROOT')
root_config.addOption('location', '%s/ROOT/6.04.02/x86_64-slc6-gcc48-opt' % getLCGRootPath(), 'Location of ROOT')
root_config.addOption('path', '', 'Set to a specific ROOT version. Will override other options.')
## Doesn't appear to work see above ^^^
#config.addOption('pythonhome', '${lcgpath}/Python/${pythonversion}/${arch}/','Location of the python used for execution of PyROOT script')
root_config.addOption('pythonhome', '%s/Python/2.7.9.p1/x86_64-slc6-gcc48-opt' % getLCGRootPath(), 'Location of the python used for execution of PyROOT script')
root_config.addOption('pythonversion', '2.7.9.p1', "Version number of python used for execution python ROOT script")
root_config.addOption('version', '6.04.02', 'Version of ROOT')

# ------------------------------------------------
# Local
local_config = makeConfig('Local', 'parameters of the local backend (jobs in the background on localhost)')
local_config.addOption('remove_workdir', True, 'remove automatically the local working directory when the job completed')
local_config.addOption('location', None, 'The location where the workdir will be created. If None it defaults to the value of $TMPDIR')

# ------------------------------------------------
# LCG
lcg_config = makeConfig('LCG', 'LCG/gLite/EGEE configuration parameters')
#gproxy_config = getConfig('GridProxy_Properties')

lcg_config.addOption('GLITE_SETUP', '/afs/cern.ch/sw/ganga/install/config/grid_env_auto.sh',
                 'sets the LCG-UI environment setup script for the GLITE middleware',
                 filter=GangaCore.Utility.Config.expandvars)

lcg_config.addOption('VirtualOrganisation', '',
                 'sets the name of the grid virtual organisation')

lcg_config.addOption('Config', '', 'sets the generic LCG-UI configuration script for the GLITE workload management system',
                 filter=GangaCore.Utility.Config.expandvars)

lcg_config.addOption(
    'AllowedCEs', '', 'sets allowed computing elements by a regular expression')
lcg_config.addOption(
    'ExcludedCEs', '', 'sets excluded computing elements by a regular expression')

lcg_config.addOption(
    'GLITE_WMS_WMPROXY_ENDPOINT', '', 'sets the WMProxy service to be contacted')
lcg_config.addOption('GLITE_ALLOWED_WMS_LIST', [], '')

lcg_config.addOption('MyProxyServer', 'myproxy.cern.ch', 'sets the myproxy server')
lcg_config.addOption('RetryCount', 3, 'sets maximum number of job retry')
lcg_config.addOption(
    'ShallowRetryCount', 10, 'sets maximum number of job shallow retry')

lcg_config.addOption(
    'Rank', '', 'sets the ranking rule for picking up computing element')
lcg_config.addOption('ReplicaCatalog', '', 'sets the replica catalogue server')
lcg_config.addOption('StorageIndex', '', 'sets the storage index')

lcg_config.addOption(
    'DefaultSE', 'srm.cern.ch', 'sets the default storage element')
lcg_config.addOption('DefaultSRMToken', '',
                 'sets the space token for storing temporary files (e.g. oversized input sandbox)')
lcg_config.addOption(
    'DefaultLFC', 'prod-lfc-shared-central.cern.ch', 'sets the file catalogue server')
lcg_config.addOption('BoundSandboxLimit', 10 * 1024 * 1024,
                 'sets the size limitation of the input sandbox, oversized input sandbox will be pre-uploaded to the storage element specified by \'DefaultSE\' in the area specified by \'DefaultSRMToken\'')

lcg_config.addOption('Requirements', 'GangaCore.Lib.LCG.LCGRequirements',
                 'sets the full qualified class name for other specific LCG job requirements')

lcg_config.addOption('SandboxCache', 'GangaCore.Lib.LCG.LCGSandboxCache',
                 'sets the full qualified class name for handling the oversized input sandbox')

lcg_config.addOption('GliteBulkJobSize', 50,
                 'sets the maximum number of nodes (i.e. subjobs) in a gLite bulk job')

lcg_config.addOption('SubmissionThread', 10,
                 'sets the number of concurrent threads for job submission to gLite WMS')

lcg_config.addOption(
    'SubmissionTimeout', 300, 'sets the gLite job submission timeout in seconds')

lcg_config.addOption('StatusPollingTimeout', 300,
                 'sets the gLite job status polling timeout in seconds')

lcg_config.addOption('OutputDownloaderThread', 10,
                 'sets the number of concurrent threads for downloading job\'s output sandbox from gLite WMS')

lcg_config.addOption('SandboxTransferTimeout', 60,
                 'sets the transfer timeout of the oversized input sandbox')

lcg_config.addOption(
    'JobLogHandler', 'WMS', 'sets the way the job\'s stdout/err are being handled.')

lcg_config.addOption('MatchBeforeSubmit', False,
                 'sets to True will do resource matching before submitting jobs, jobs without any matched resources will fail the submission')

lcg_config.addOption('IgnoreGliteScriptHeader', False,
                 'sets to True will load script-based glite-wms-* commands forcely with current python, a trick for 32/64 bit compatibility issues.')

# add ARC specific configuration options
#lcg_config.addOption('ArcInputSandboxBaseURI', '', 'sets the baseURI for getting the input sandboxes for the job')
#lcg_config.addOption('ArcOutputSandboxBaseURI', '', 'sets the baseURI for putting the output sandboxes for the job')
lcg_config.addOption('ArcWaitTimeBeforeStartingMonitoring', 240,
                 'Time in seconds to wait after submission before starting to monitor ARC jobs to ensure they are in the system')
lcg_config.addOption('ArcJobListFile', "~/.arc/gangajobs.xml",
                 'File to store ARC job info in when submitting and monitoring, i.e. argument to "-j" option in arcsub. Ganga default is different to ARC default (~/.arc/jobs.xml) to keep them separate.')
lcg_config.addOption('ArcConfigFile', "",
                 'Config file for ARC submission. Use to specify CEs, etc. Default is blank which will mean no config file is specified and the default (~/arc/client.conf) is used')
lcg_config.addOption('ArcCopyCommand', 'arcget',
                  'sets the copy command for ARC when dealing with sandboxes')
#lcg_config.addOption('ArcPrologue','','sets the prologue script')
#lcg_config.addOption('ArcEpilogue','','sets the epilogue script')

# add CREAM specific configuration options
lcg_config.addOption('CreamInputSandboxBaseURI', '',
                 'sets the baseURI for getting the input sandboxes for the job')
lcg_config.addOption('CreamOutputSandboxBaseURI', '',
                 'sets the baseURI for putting the output sandboxes for the job')
lcg_config.addOption('CreamCopyCommand', 'gfal-copy-url',
                 'sets the copy command for CREAM when dealing with sandboxes')
#lcg_config.addOption('CreamPrologue','','sets the prologue script')
#lcg_config.addOption('CreamEpilogue','','sets the epilogue script')

# ------------------------------------------------
# GridSimulator
gridsim_config = makeConfig('GridSimulator', 'Grid Simulator configuration parameters')

gridsim_config.addOption('submit_time', 'random.uniform(1,10)',
                 'python expression which returns the time it takes (in seconds) to complete the Grid.submit() command (also for subjob in bulk emulation)')
gridsim_config.addOption(
    'submit_failure_rate', 0.0, 'probability that the Grid.submit() method fails')

gridsim_config.addOption('cancel_time', 'random.uniform(1,5)',
                 'python expression which returns the time it takes (in seconds) to complete the Grid.cancel() command (also for subjob in bulk emulation)')
gridsim_config.addOption(
    'cancel_failure_rate', 0.0, 'probability that the Grid.cancel() method fails')

gridsim_config.addOption('status_time', 'random.uniform(1,5)',
                 'python expression which returns the time it takes (in seconds) to complete the status command (also for subjob in bulk emulation)')

gridsim_config.addOption('get_output_time', 'random.uniform(1,5)',
                 'python expression which returns the time it takes (in seconds) to complete the get_output command (also for subjob in bulk emulation)')

#config.addOption('bulk_submit_time','random.uniform(1,2)','python expression which returns the time it takes (in seconds) to complete the submission of a single job within the Grid.native_master_submit() command')
#config.addOption('bulk_submit_failure_rate',0.0,'probabilty that the Grid.native_master_submit() fails')

#config.addOption('bulk_cancel_time','random.uniform(1,2)','python expression which returns the time it takes (in seconds) to complete the cancellation of a single job within the Grid.native_master_cancel() command')
#config.addOption('bulk_cancel_failure_rate',0.0,'probabilty that the Grid.native_master_cancel() fails')

gridsim_config.addOption('job_id_resolved_time', 'random.uniform(1,2)',
                 'python expression which returns the time it takes (in seconds) to complete the resolution of all the id of a subjob (when submitted in bulk) this is the time the NODE_ID becomes available from the monitoring)')

#config.addOption('job_scheduled_time','random.uniform(10,20)', 'python expression which returns the time the job stays in the scheduled state')
#config.addOption('job_running_time','random.uniform(10,20)', 'python expression which returns the time the job stays in the running state')
gridsim_config.addOption('job_finish_time', 'random.uniform(10,20)',
                 'python expression which returns the time when the job enters the Done success or Failed state')
gridsim_config.addOption(
    'job_failure_rate', 0.0, 'probability of the job to enter the Failed state')

# ------------------------------------------------
# Condor
condor_config = makeConfig('Condor', 'Settings for Condor Batch system')

condor_config.addOption('query_global_queues', True,
                 "Query global condor queues, i.e. use '-global' flag")

# ------------------------------------------------
# LSF
lsf_config = makeConfig('LSF', 'internal LSF command line interface')

# fix bug #21687
lsf_config.addOption('shared_python_executable', False, "Shared PYTHON")

lsf_config.addOption('jobid_name', 'LSB_BATCH_JID', "Name of environment with ID of the job")
lsf_config.addOption('queue_name', 'LSB_QUEUE', "Name of environment with queue name of the job")
lsf_config.addOption('heartbeat_frequency', '30', "Heartbeat frequency config variable")

lsf_config.addOption('submit_str', 'cd %s; bsub %s %s %s %s', "String used to submit job to queue")
lsf_config.addOption('submit_res_pattern', '^Job <(?P<id>\d*)> is submitted to .*queue <(?P<queue>\S*)>',
                 "String pattern for replay from the submit command")

lsf_config.addOption('stdoutConfig', '-o %s/stdout', "String pattern for defining the stdout")
lsf_config.addOption('stderrConfig', '-e %s/stderr', "String pattern for defining the stderr")

lsf_config.addOption('kill_str', 'bkill %s', "String used to kill job")
lsf_config.addOption('kill_res_pattern',
                 '(^Job <\d+> is being terminated)|(Job <\d+>: Job has already finished)|(Job <\d+>: No matching job found)',
                 "String pattern for replay from the kill command")

tempstr = '''
'''
lsf_config.addOption('preexecute', tempstr,
                 "String contains commands executing before submiting job to queue")

tempstr = '''
def filefilter(fn):
  # FILTER OUT Batch INTERNAL INPUT/OUTPUT FILES:
  # 10 digits . any number of digits . err or out
  import re
  internals = re.compile(r'\d{10}\.\d+.(out|err)')
  return internals.match(fn) or fn == '.Batch.start'
'''
lsf_config.addOption('postexecute', tempstr, "String contains commands executing before submiting job to queue")
lsf_config.addOption('jobnameopt', 'J', "String contains option name for name of job in batch system")
lsf_config.addOption('timeout', 600, 'Timeout in seconds after which a job is declared killed if it has not touched its heartbeat file. Heartbeat is touched every 30s so do not set this below 120 or so.')

# ------------------------------------------------
# PBS
pbs_config = makeConfig('PBS', 'internal PBS command line interface')

pbs_config.addOption('shared_python_executable', False, "Shared PYTHON")

pbs_config.addOption('jobid_name', 'PBS_JOBID', "Name of environment with ID of the job")
pbs_config.addOption('queue_name', 'PBS_QUEUE', "Name of environment with queue name of the job")
pbs_config.addOption('heartbeat_frequency', '30', "Heartbeat frequency config variable")

pbs_config.addOption('submit_str', 'cd %s; qsub %s %s %s %s', "String used to submit job to queue")
pbs_config.addOption('submit_res_pattern', '^(?P<id>\d*)\.pbs\s*',
                 "String pattern for replay from the submit command")

pbs_config.addOption('stdoutConfig', '-o %s/stdout', "String pattern for defining the stdout")
pbs_config.addOption('stderrConfig', '-e %s/stderr', "String pattern for defining the stderr")

pbs_config.addOption('kill_str', 'qdel %s', "String used to kill job")
pbs_config.addOption('kill_res_pattern', '(^$)|(qdel: Unknown Job Id)',
                 "String pattern for replay from the kill command")

tempstr = '''
env = os.environ
jobnumid = env["PBS_JOBID"]
os.system("mkdir /tmp/%s/" %jobnumid)
os.chdir("/tmp/%s/" %jobnumid)
os.environ["PATH"]+=":."
'''
pbs_config.addOption('preexecute', tempstr,
                 "String contains commands executing before submiting job to queue")

tempstr = '''
env = os.environ
jobnumid = env["PBS_JOBID"]
os.chdir("/tmp/")
os.system("rm -rf /tmp/%s/" %jobnumid)
'''
pbs_config.addOption('postexecute', tempstr,
                 "String contains commands executing before submiting job to queue")
pbs_config.addOption('jobnameopt', 'N', "String contains option name for name of job in batch system")
pbs_config.addOption('timeout', 600,
                 'Timeout in seconds after which a job is declared killed if it has not touched its heartbeat file. Heartbeat is touched every 30s so do not set this below 120 or so.')

# ------------------------------------------------
# SGE
sge_config = makeConfig('SGE', 'internal SGE command line interface')

sge_config.addOption('shared_python_executable', False, "Shared PYTHON")

sge_config.addOption('jobid_name', 'JOB_ID', "Name of environment with ID of the job")
sge_config.addOption('queue_name', 'QUEUE', "Name of environment with queue name of the job")
sge_config.addOption('heartbeat_frequency', '30', "Heartbeat frequency config variable")

# the -V options means that all environment variables are transferred to
# the batch job (ie the same as the default behaviour on LSF at CERN)
sge_config.addOption('submit_str', 'cd %s; qsub -cwd -S /usr/bin/python -V %s %s %s %s',
                 "String used to submit job to queue")
sge_config.addOption('submit_res_pattern', 'Your job (?P<id>\d+) (.+)',
                 "String pattern for replay from the submit command")

sge_config.addOption('stdoutConfig', '-o %s/stdout', "String pattern for defining the stdout")
sge_config.addOption('stderrConfig', '-e %s/stderr', "String pattern for defining the stderr")

sge_config.addOption('kill_str', 'qdel %s', "String used to kill job")
sge_config.addOption('kill_res_pattern', '(has registered the job +\d+ +for deletion)|(denied: job +"\d+" +does not exist)',
                 "String pattern for replay from the kill command")

# From the SGE man page on qsub
#
#===========================
# Furthermore, Grid Engine sets additional variables into the job's
# environment, as listed below.
#:
#:
# TMPDIR
#   The absolute path to the job's temporary working directory.
#=============================


sge_config.addOption('preexecute', 'os.chdir(os.environ["TMPDIR"])\nos.environ["PATH"]+=":."',
                 "String contains commands executing before submiting job to queue")
sge_config.addOption('postexecute', '', "String contains commands executing before submiting job to queue")
sge_config.addOption('jobnameopt', 'N', "String contains option name for name of job in batch system")
sge_config.addOption('timeout', 600, 'Timeout in seconds after which a job is declared killed if it has not touched its heartbeat file. Heartbeat is touched every 30s so do not set this below 120 or so.')

# ------------------------------------------------
# Slurm
slurm_config = makeConfig('Slurm', 'internal Slurm command line interface')

slurm_config.addOption('shared_python_executable', False, "Shared PYTHON")

slurm_config.addOption('jobid_name', 'SLURM_JOB_ID', "Name of environment with ID of the job")
# Note: SLURM queues are called partitions.
slurm_config.addOption('queue_name', 'SLURM_JOB_PARTITION', "Name of environment with partition name of the job")
slurm_config.addOption('heartbeat_frequency', '30', "Heartbeat frequency config variable")

slurm_config.addOption('submit_str', 'cd %s; sbatch %s %s %s %s', "String used to submit job to partition")
slurm_config.addOption('submit_res_pattern', '^Submitted batch job (?P<id>\d+)\s*',
                       "String pattern for replay from the submit command")

slurm_config.addOption('stdoutConfig', '-o %s/stdout', "String pattern for defining the stdout")
slurm_config.addOption('stderrConfig', '-e %s/stderr', "String pattern for defining the stderr")

slurm_config.addOption('kill_str', 'scancel %s', "String used to kill job")
slurm_config.addOption('kill_res_pattern', '(^$)|(^scancel: error: .+)',
                       "String pattern for replay from the kill command")

#  Note: some SLURM systems automatically set the TMPDIR environment
#        variable, which points to the job's temporary directory.
#        Make sure that you remove all created files from there, otherwise
#        SLURM will leave this directory (together with all files).
#  Warning: some SLURM systems set TMPDIR="/tmp" and then the SLURM_TMPDIR
#        environment variable possibly points to the job's temporary directory.
#  Note: some SLURM systems automatically set the SCRATCHDIR environment
#        variable, which points to the job's scratch directory.
#        Make sure that you remove all created files from there, otherwise
#        SLURM will leave this directory (together with all files).
#        Usually TMPDIR="${SCRATCHDIR}/tmp"
#  Note: some SLURM systems can set the LOCALFS and / or the SCRATCH_LOCAL
#        environment variables, which point to the job's temporary directory
#        and which will automatically disappear when the job ends.
#  Note: some SLURM systems can set the MEMFS environment variable, which
#        points to the job's "RAM disk" resident temporary directory
#        and which will automatically disappear when the job ends.
#  Note: some (all?) computer clusters set the SCRATCH environment variable.
#        Usually SCRATCHDIR="${SCRATCH}/slurm_jobdir/${SLURM_JOB_ID}"

tempstr = '''
env = os.environ
jobnumid = env.get("SLURM_JOB_ID") or env.get("SLURM_JOBID") or "pid_"+str(os.getpid())
scratchDir = env.get("MEMFS") or env.get("LOCALFS") or env.get("SCRATCH_LOCAL") or env.get("SCRATCHDIR") or env.get("SLURM_TMPDIR") or env.get("SCRATCH") or env.get("TMPDIR") or "/tmp"
scratchDir = scratchDir+"/workdir"
if not jobnumid in scratchDir: scratchDir = scratchDir+"_"+jobnumid
os.system("mkdir -p "+scratchDir)
os.chdir(scratchDir)
# env["PATH"]+=":."
'''
slurm_config.addOption('preexecute', tempstr,
                       "String contains the first commands executing right after the job starts")

tempstr = '''
env = os.environ
jobnumid = env.get("SLURM_JOB_ID") or env.get("SLURM_JOBID") or "pid_"+str(os.getpid())
scratchDir = env.get("MEMFS") or env.get("LOCALFS") or env.get("SCRATCH_LOCAL") or env.get("SCRATCHDIR") or env.get("SLURM_TMPDIR") or env.get("SCRATCH") or env.get("TMPDIR") or "/tmp"
scratchDir = scratchDir+"/workdir"
if not jobnumid in scratchDir: scratchDir = scratchDir+"_"+jobnumid
os.chdir("/tmp/")
os.system("rm -rf "+scratchDir)
'''
slurm_config.addOption('postexecute', tempstr,
                       "String contains the last commands executing right before the job ends")
slurm_config.addOption('jobnameopt', 'J', "String contains option name for name of job in batch system")
slurm_config.addOption('timeout', 600,
                       'Timeout in seconds after which a job is declared killed if it has not touched its heartbeat file. Heartbeat is touched every 30s so do not set this below 120 or so.')

# ------------------------------------------------
# Mergers
merge_config = makeConfig('Mergers', 'parameters for mergers')
merge_config.addOption('associate', {'log': 'TextMerger', 'root': 'RootMerger',
                               'text': 'TextMerger', 'txt': 'TextMerger'}, 'Dictionary of file associations')
gangadir = getConfig('Configuration')['gangadir']
merge_config.addOption('merge_output_dir', gangadir +
                 '/merge_results', "location of the merger's outputdir")
merge_config.addOption('std_merge', 'TextMerger', 'Standard (default) merger')

# ------------------------------------------------
# Preparable
preparable_config = makeConfig('Preparable', 'Parameters for preparable applications')
preparable_config.addOption('unprepare_on_copy', False, 'Unprepare a prepared application when it is copied')

# ------------------------------------------------
# GPIComponentFilters
gpicomp_config = makeConfig('GPIComponentFilters', """Customization of GPI component object assignment
for each category there may be multiple filters registered, the one used being defined
in the configuration file in [GPIComponentFilters]
e.g: {'datasets':{'lhcbdatasets':lhcbFilter, 'testdatasets':testFilter}...}
""", is_open=False)

# ------------------------------------------------
# Output
output_config = makeConfig("Output", "configuration section for postprocessing the output")
output_config.addOption('AutoRemoveFilesWithJob', False,
                       'if True, each outputfile of type in list AutoRemoveFileTypes will be removed when the job is')
output_config.addOption('AutoRemoveFileTypes', [
                       'DiracFile'], 'List of outputfile types that will be auto removed when job is removed if AutoRemoveFilesWithJob is True')

output_config.addOption('PostProcessLocationsFileName', '__postprocesslocations__',
                       'name of the file that will contain the locations of the uploaded from the WN files')

output_config.addOption('FailJobIfNoOutputMatched', True,
                       'if True, a job will be marked failed if output is asked for but not found.')

output_config.addOption('ForbidLegacyOutput', True, 'if True, writing to the job outputdata and outputsandbox fields will be forbidden')

output_config.addOption('ForbidLegacyInput', True, 'if True, writing to the job inputsandbox field will be forbidden')

docstr_Ext = 'fileExtensions:list of output files that will be written to %s,\
backendPostprocess:defines where postprocessing should be done (WN/client) on different backends,\
uploadOptions:config values needed for the actual %s upload'

# LocalFile
LocalPost = {'Local': 'client',
             'Interactive': 'client',
             'LSF': 'client',
             'SGE': 'client',
             'PBS': 'client',
             'Condor': 'client',
             'CREAM': 'client',
             'ARC': 'client',
             'Dirac': 'client'
}
LocalUpOpt = {}
LocalFileExt = docstr_Ext % ('Local', 'Local')
output_config.addOption('LocalFile',
                           {'fileExtensions': ['*.txt'],
                             'backendPostprocess' : LocalPost,
                             'uploadOptions' : LocalUpOpt},
                            LocalFileExt)


# LCGSEFILE

LCGSEBakPost = {'LSF': 'client',
                'PBS': 'client',
                'SGE': 'client',
                'Condor': 'client',
                'LCG': 'WN',
                'CREAM': 'WN',
                'ARC': 'WN',
                'Local': 'WN',
                'Interactive': 'WN'
}
LCGSEUpOpt = {'LFC_HOST': 'lfc-dteam.cern.ch', 'dest_SRM': 'srm-public.cern.ch'}
LCGSEFileExt = docstr_Ext % ('LCG SE', 'LCG')

output_config.addOption('LCGSEFile',
                       {'fileExtensions': ['*.root', '*.asd'],
                        'backendPostprocess': LCGSEBakPost,
                        'uploadOptions': LCGSEUpOpt},
                       LCGSEFileExt)

# DiracFile
## TODO MOVE ME TO GANGADIRAC!!!
# Should this be in Core or elsewhere?
diracBackPost = {'Dirac': 'submit',
                 'LSF': 'WN',
                 'PBS': 'WN',
                 'SGE': 'WN',
                 'Condor': 'WN',
                 'LCG': 'WN',
                 'CREAM': 'WN',
                 'ARC': 'WN',
                 'Local': 'WN',
                 'Interactive': 'WN'}
diracFileExts = docstr_Ext % ('DIRAC', 'DIRAC')

output_config.addOption('DiracFile',
                       {'fileExtensions': ['*.dst'],
                        'backendPostprocess': diracBackPost,
                        'uploadOptions': {},
                        'defaultSite': {'upload': 'CERN-USER', 'download': 'CERN-USER'}},
                       diracFileExts)

# GoogleFile

GoogleFileBackPost = {'Dirac': 'client',
                      'LSF': 'client',
                      'PBS': 'client',
                      'SGE': 'client',
                      'Condor': 'client',
                      'LCG': 'client',
                      'CREAM': 'client',
                      'ARC': 'client',
                      'Local': 'client',
                      'Interactive': 'client'}
GoogleFileExts = docstr_Ext % ('GoogleDrive', 'Google')

output_config.addOption('GoogleFile',
                       {'fileExtensions': [],
                        'backendPostprocess': GoogleFileBackPost,
                        'uploadOptions': {}},
                       GoogleFileExts)

# MassStorageFile

import pwd
import grp
Conf_config = getConfig('Configuration')
if 'user' in Conf_config:
    user = Conf_config ['user']
else:
    #import sys
    #sys.stderr.write('Configure Error: %s' % str(err) )
    import getpass
    user = getpass.getuser()


## FIXME Sometimes the wrong user is set gere for the unittests, I've added this to correct for it - rcurrie
try:
    pwd_nam = pwd.getpwnam(user)
except:
    import getpass
    user = getpass.getuser()
    pwd_nam = pwd.getpwnam(user)

groupid = grp.getgrgid(pwd_nam.pw_gid).gr_name
groupnames = {'z5': 'lhcb', 'zp': 'atlas', 'zh': 'cms', 'vl': 'na62'}
groupname = groupnames.get(groupid, 'undefined')

try:
    import os.path
    massStoragePath = os.path.join(os.environ['EOS_HOME'], 'ganga')
except KeyError:
    massStoragePath = "/eos/%s/user/%s/%s/ganga" % (
        groupname, user[0], user)

# From:
# http://eos.cern.ch/index.php?option=com_content&view=article&id=87:using-eos-at-cern&catid=31:general&Itemid=41
protoByExperiment = {'atlas': 'root://eosatlas.cern.ch',
                     'cms': 'root://eocms.cern.ch',
                     'lhcb': 'root://eoslhcb.cern.ch',
                     'alice': 'root://eosalice.cern.ch',
                     # These last 2 are guesses based on the standard
                     'na62': 'root://eosna62.cern.ch',
                     'undefined': 'root://eos.cern.ch'}
defaultMassStorageProto = protoByExperiment[groupname]

eosinstalled, prefix = commands.getstatusoutput('which eos')
if not eosinstalled == 0:
    prefix = ''
    massStoragePath = ''    

massStorageUploadOptions = {'mkdir_cmd': prefix + ' mkdir', 'cp_cmd':
                            prefix + ' cp', 'ls_cmd': prefix + ' ls', 'path': massStoragePath}

massStorageFileExt = docstr_Ext % ('Mass Storage', 'EOS')

massStorageBackendPost = {'LSF': 'WN',
                          'PBS': 'WN',
                          'Condor': 'WN',
                          'SGE': 'WN',
                          'LCG': 'client',
                          'CREAM': 'client',
                          'ARC': 'client',
                          'Local': 'WN',
                          'Interactive': 'client',
                          'Dirac': 'client'}

output_config.addOption('MassStorageFile',
                       {'fileExtensions': [''],
                        'backendPostprocess': massStorageBackendPost,
                        'uploadOptions': massStorageUploadOptions,
                        'defaultProtocol': defaultMassStorageProto},
                       massStorageFileExt)

sharedFileBackendPost = {'LSF': 'WN',
                         'LCG': 'client',
                         'ARC': 'client',
                         'Dirac': 'client',
                         'PBS': 'WN',
                         'SGE': 'WN',
                         'Condor': 'WN',
                         'Interactive': 'client',
                         'Local': 'WN',
                         'CREAM': 'client'}

output_config.addOption('SharedFile',
                       {'fileExtensions': [''],
                        'backendPostprocess': sharedFileBackendPost,
                        'uploadOptions': {'path': None, 'cp_cmd': 'cp', 'ls_cmd': 'ls', 'mkdir_cmd': 'mkdir'},
                        'defaultProtocol': 'file://'},
                        docstr_Ext % ('Shared Storage', 'SharedFS'))

# ------------------------------------------------
# Display
disp_config = makeConfig('Display', 'control the printing style of the different registries ("jobs","box","tasks"...)')
disp_config.addOption('config_name_colour', 'fx.bold',
                         'colour print of the names of configuration sections and options')
disp_config.addOption(
    'config_docstring_colour', 'fg.green', 'colour print of the docstrings and examples')
disp_config.addOption(
    'config_value_colour', 'fx.bold', 'colour print of the configuration values')
disp_config.addOption('jobs_columns',
                 ("fqid", "status", "name", "subjobs", "application",
                  "backend", "backend.actualCE", "comment", "subjob status"),
                 'list of job attributes to be printed in separate columns')

disp_config.addOption('jobs_columns_width',
                 {'fqid': 8, 'status': 10, 'name': 10, 'subjobs': 8, 'application':
                     15, 'backend': 15, 'backend.actualCE': 45, 'comment': 30, 'subjob status': 15},
                 'width of each column')

disp_config.addOption('jobs_columns_functions',
                 {'subjobs': "lambda j: len(j.subjobs)", 'application': "lambda j: j.application.__class__.__name__",
                  'backend': "lambda j:j.backend.__class__.__name__", 'comment': "lambda j: j.comment", 'subjob status': "lambda j: j.returnSubjobStatuses()"},
                 'optional converter functions')

disp_config.addOption('jobs_columns_show_empty',
                 ['fqid'],
                 'with exception of columns mentioned here, hide all values which evaluate to logical false (so 0,"",[],...)')

disp_config.addOption('jobs_status_colours',
                 {'new': 'fx.normal',
                  'submitted': 'fg.orange',
                  'running': 'fg.green',
                  'completed': 'fg.blue',
                  'failed': 'fg.red'
                  },
                 'colours for jobs status')

# add display default values for the box
disp_config.addOption('box_columns',
                 ("id", "type", "name", "application"),
                 'list of job attributes to be printed in separate columns')

disp_config.addOption('box_columns_width',
                 {'id': 5, 'type': 20, 'name': 40, 'application': 15},
                 'width of each column')

disp_config.addOption('box_columns_functions',
                 {'application': "lambda obj: obj.application._name"},
                 'optional converter functions')

disp_config.addOption('box_columns_show_empty',
                 ['id'],
                 'with exception of columns mentioned here, hide all values which evaluate to logical false (so 0,"",[],...)')

# display default values for task list
markup = ANSIMarkup()
str_done = markup("done", overview_colours["completed"])
disp_config.addOption('tasks_columns',
                     ("id", "Type", "Name", "State",
                      "Comment", "Jobs", str_done),
                     'list of job attributes to be printed in separate columns')

disp_config.addOption('tasks_columns_width',
                     {"id": 5, "Type": 13, "Name": 22, "State": 9,
                         "Comment": 30, "Jobs": 33, str_done: 5},
                     'width of each column')

disp_config.addOption('tasks_columns_functions',
                     {'Name': "lambda t : t.name",
                      'Type': "lambda task : task._name",
                      'State ': "lambda task : task.status",
                      'Comment ': "lambda task : task.comment",
                      'Jobs': "lambda task : task.n_all()",
                      str_done: "lambda task : task.n_status('completed')",
                      },
                     'optional converter functions')

disp_config.addOption('tasks_columns_show_empty',
                     ['id', 'Jobs',
                         str_done],
                     'with exception of columns mentioned here, hide all values which evaluate to logical false (so 0,"",[],...)')

disp_config.addOption(
    'tasks_show_help', True, 'change this to False if you do not want to see the help screen if you first type "tasks" in a session')

# ------------------------------------------------
# Tasks
tasks_config = makeConfig('Tasks', 'Tasks configuration options')
tasks_config.addOption('TaskLoopFrequency', 60., "Frequency of Task Monitoring loop in seconds")
tasks_config.addOption('ForceTaskMonitoring', False, "Monitor tasks even if the monitoring loop isn't enabled")
tasks_config.addOption('disableTaskMon', False, "Should I disable the Task Monitoring loop?")

# ------------------------------------------------
# MonitoringServices
mon_config = makeConfig('MonitoringServices', """External monitoring systems are used
to follow the submission and execution of jobs. Each entry in this section
defines a monitoring plugin used for a particular combination of application
and backend. Asterisks may be used to specify any application or any
backend. The configuration entry syntax:

ApplicationName/BackendName = dot.path.to.monitoring.plugin.class.

Example: DummyMS plugin will be used to track executables run on all backends:

Executable/* = GangaCore.Lib.MonitoringServices.DummyMS.DummyMS

""", is_open=True)

# ------------------------------------------------
# Registry Dirty Monitoring Services (not related to actual Job Monitoring)
reg_config = makeConfig('Registry','This config controls the speed of flushing objects to disk')
reg_config.addOption('AutoFlusherWaitTime', 30, 'Time to wait between auto-flusher runs')
reg_config.addOption('EnableAutoFlush', True, 'Enable Registry auto-flushing feature')
reg_config.addOption('DisableLoadCheck', True, 'Disable the checking of recent bad jobs in bad state. Mainly used in testing.')

cred_config = makeConfig('Credentials', 'This configures the credentials singleton')
cred_config.addOption('CleanDelay', 1, 'Seconds between auto-clean of credentials when proxy externally destroyed')
cred_config.addOption('AtomicDelay', 1, 'Seconds between checking credential on disk')



#!/usr/bin/env ganga

import re
import inspect
import GangaPlotHelper

## require matplotlib to produce statistic plots
from pylab import *

import sys
from functools import reduce

from GangaCore.Utility.Config import makeConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.util import isStringLike, canLoopOver

logger = getLogger()
config = makeConfig('GangaPlotter','GangaPlotter configuration parameters')

NOT_APPLICABLE = '*N/A*'

class GetJobAttrException(Exception):

    '''exception for parser error'''

    def __init__(self,msg):
        self.msg = msg
    def __str__(self):
        return self.msg

## this function could be put in GangaCore.Utility.util
def isNumeric(myValue):
    '''checks via type() if myVar is numeric or not'''
    if type(myValue)==type(1) or type(myValue)==type(1) or \
       type(myValue)==type(1.1) or type(myValue)==type(1+1j) :
       return 1
    else:
       return 0

def __getJobAttribute__(job,attrName):

    '''method for getting the value of the given attribute from the GANGA job object'''

    attrList  = attrName.split('.')
    attrValue = job

    for attr in attrList:
        if not re.match(r'^\_\_.*\_\_$',attr):
            try:
                attrValue = getattr(attrValue,attr)
            except Exception as x:
                errMsg = 'Cannot select attribute %s: %s!' % (attr,str(x))
                raise GetJobAttrException(errMsg)
        else:
            errMsg = 'Query the attribute %s is not allowed!' % attr 
            raise GetJobAttrException(errMsg)

    attrClass = attrValue.__class__.__name__

    try:
        attrValue = attrValue._impl._name
    except AttributeError:
        pass
    
    return (attrValue, attrClass)

def __makeJobInfo__(job,attrNameList):

    '''make complete jobinfo according to the given list of job attributs'''

    jobInfo = []
    for attrName in attrNameList:
        try:
            if isStringLike(attrName):
                attrValue, attrClass = __getJobAttribute__(job,attrName)
            else:
                attrValue, attrClass = attrName(job), None
            jobInfo.append(attrValue)
        except Exception as x:
            logger.warning('Ignoring job %d: %s'%(job.id,str(x)))
            jobInfo.append(NOT_APPLICABLE)

    return jobInfo 

def __makePlotTitle__(num_jobs,deep,subtitle):

    title = 'Statistics for %d jobs ' % num_jobs

    if(deep):
        title += ' (w/ subjobs): '
    else:
        title += ' (w/o subjobs): '

    return title+subtitle

def getJobInfoTable(jobs,attrNameList,deep=True):

    '''prepare the data table for plots'''
    
    jobInfoTable = []
    jobInfoTable.append(attrNameList)
    id = 0
    try:
        for j in jobs:
            if deep and len(j.subjobs) > 0:
                # get the infotable from subjobs
                # The deep looping will go through all of the subjob levels
                info = getJobInfoTable(j.subjobs,attrNameList,deep)

                # extend the main infotable to include the infotable of the subjobs 
                # the title row should be substracted before merging
                jobInfoTable.extend(info[1:])
                id = id + len(info[1:])
            else:
                # if deep looping is off or there is no subjob inside
                # take the info directly from the job 
                jobInfoTable.append(__makeJobInfo__(j,attrNameList))
                id += 1

    except Exception as e:
        raise e

    return jobInfoTable


class GangaPlotter:

    ''' the object of GANGA job statistical plots '''

    def __init__(self):
        self.figId  = 0

    def __setFigureId__(self):
        self.figId += 1

    def __doPlot__(self):            
        ## savefile if requested
        if self.output is not None:
            savefig(self.output)
            logger.info('the plot has been saved in %s' % self.output)

        ## force to plot if interactive mode is not activated
        if not isinteractive():
            show()
            close()

    def __unionLists__(self,lists):

        unionSet = set(lists[0])

        for i in range(1,len(lists)):
            newSet   = set(lists[i])
            unionSet.update(newSet)

        unionList = []
        for i in range(len(unionSet)):
            unionList.append(unionSet.pop())

        return unionList

    def __makeList__(self,element):

        if isStringLike(element):
            myList = [element]
        elif canLoopOver(element):
            myList = element
        else:
            myList = [element]
        return myList

    def __setDataProcessor__(self,attr,attrext,dataproc):

        """ set the data processor for attribute values """

        if attrext is not None and dataproc is not None:
            logger.info('user defined dataproc %s will be used.' % str(dataproc))
        elif attr in ['backend.CE','backend.actualCE']:
            # the build-in data processors 
            if attrext == 'by_queue':
                dataproc = None
            if attrext == 'by_ce':
                dataproc = lambda x:x.split(':2119/')[0]
            if attrext == 'by_country':
                dataproc = lambda x:GangaPlotHelper.tld_country(x.split(':2119/')[0].split('.')[-1])
        return dataproc

    def __setColormap__(self, pltColorMap, numColors):

        """ setup the color map """

        plt_cmap = []
        if not pltColorMap is None:
            ## load user defined pltColorMap
            logger.info('load user defined color map.')
            for i in range(numColors):
                plt_cmap.append(pltColorMap[i % len(pltColorMap)])
        else:
            #plt_cmap = self.__defaultColormap__(numColors)
            plt_cmap = self.__defaultColormapAlt__(numColors)

        return plt_cmap

    def __defaultColormapAlt__(self,dataBlockNumber):

        '''the default color map definition'''

        c_min = 0.1
        c_max = 0.9
        c_delta_mag = 1.0*(c_max - c_min) / dataBlockNumber

        c_map = []
        for i in range(dataBlockNumber):
            c_mag = c_min + i * c_delta_mag
            c_map.append(GangaPlotHelper.floatRgb(c_mag,c_min,c_max))

        return c_map

    def __defaultColormap__(self,dataBlockNumber):

        '''the default color map definition'''
 
        c_i_min = 400 - 200
        c_i_max = 400 + 100
 
        c_map = []
 
        for i in range(dataBlockNumber):
            ## select the corresponding color for displaying the fraction
            c_i = i % (c_i_max-c_i_min+1)
            if i % 2 == 0:
                c_i = (c_i + c_i_min)
            else:
                c_i = (c_i_max - i)
            c_x = 1.0 * int(c_i/64)
            c_y = 1.0 * int((c_i % 64)/8)
            c_z = 1.0 * c_i - 64*c_x - 8*c_y
            color = (c_x/8.0,c_y/8.0,c_z/8.0)
 
            c_map.append(color)
 
        return c_map

    def __defaultSubtitle__(self,titleAttr):

        """ compose the subtitle according to the given titleAttr. """

        if isStringLike(titleAttr):  ## a string
            return titleAttr
        elif canLoopOver(titleAttr): ## a list
            return self.__defaultSubtitle__(titleAttr[0])
        elif inspect.isfunction(titleAttr): ## other things: function, object ...
            return 'user function %s' % titleAttr.__name__

    def __procPltData__(self,value,pltDataProc):

        """ data processing on the attribute value """

        if not pltDataProc is None:
            try:
                value = pltDataProc(value)
            except Exception as x:
                logger.warning('Ignoring data processing on %s: %s'%(str(value),str(x)))
        else:
            value = value

        return value

    def __makeScatter__(self,dataTable,pltXColId=0,pltYColId=1,pltTitle='Scatter Plot',pltXLabel=None,pltYLabel=None,pltColorMap=None,pltOutput=None,pltXDataProc=None,pltYDataProc=None,pltCDataProc=None):

        """backend scatter plot generator """

        ## the last marker is a ntuple for "star"
        markerList = ['o','s','^','d','+','v','x', (5,1,0)]

        pltAlpha=0.7

        dataHeader = dataTable[0]
        dataTable  = dataTable[1:]

        if len(dataTable) < 1:
            logger.warning('Scatter plot requires at least 2 columns in the data Table, the given contains only %d.') % len(dataTable)
            return

        xdata = {} 
        ydata = {} 

        i = 0
        for data in dataTable:

            logger.info('processing data row %d' % i)

            i+=1

            xval = self.__procPltData__(data[pltXColId],pltXDataProc)
            yval = self.__procPltData__(data[pltYColId],pltYDataProc)
            
            if len(data) >= 3:
                cval = self.__procPltData__(data[2],pltCDataProc)
            else:
                cval = pltTitle

            if cval not in xdata.keys():
               xdata[cval] = []
               ydata[cval] = []

            xdata[cval].append(xval)
            ydata[cval].append(yval)       

        if not pltXLabel:
            pltXLabel = dataHeader[pltXColId]

        if not pltYLabel:
            pltYLabel = dataHeader[pltYColId]

        ##-----
        ## Generating marker and color array
        ##-----
        legend_labels = xdata.keys()
        mycmap = self.__setColormap__( pltColorMap, len(legend_labels) )

        ##-----
        ## Generating the scatter plot
        ##-----
        self.output = pltOutput
        self.__setFigureId__()

        figure(self.figId)
        rc('font',size=8.0)

        i = 0

        charts = []
        for key in legend_labels:

            marker = markerList[ i % len(markerList) ]
            chart = scatter(xdata[key], ydata[key], c=[ mycmap[i] ]*len(xdata[key]), cmap=mycmap, marker=marker, alpha=pltAlpha, label=key)
            charts.append(chart)

            i += 1

        grid(True)
        title(pltTitle)
        ylabel(pltYLabel)
        xlabel(pltXLabel)
        #legend(shadow=True, loc='best')
        legend(charts, legend_labels, shadow=True, loc='best')
        axis('on')

        self.__doPlot__()

    def __makeBarChart__(self,dataTable,pltXColId=0,pltYColIds=[1],pltTitle='Bar Chart',pltXLabel=None,pltYLabel='#',pltColorMap=None,pltOutput=None,pltXDataProc=None,pltYDataProcs=None,stackedBar=True):

        """ backend bar chart generator """

        dataHeader = dataTable[0]
        dataTable  = dataTable[1:]

        if len(dataTable) < 1:
            logger.warning('Barchart requires at least 2 columns in the data Table, the given contains only %d.') % len(dataTable)
            return

        pltYColIds    = self.__makeList__(pltYColIds)
        pltYDataProcs = self.__makeList__(pltYDataProcs)

        for i in range(len(pltYDataProcs),len(pltYColIds)):
            pltYDataProcs.append(None)

        ##----- 
        ## extract the given data for pie chart 
        ##----- 
        # summarize the data of the given column 
        pltData = {}
        for data in dataTable:
            # apply the user defined data processor
            xLabel = self.__procPltData__(data[pltXColId],pltXDataProc)
            # should make sure all the values are represented in String type
            xLabel  = str(xLabel)

            for i in range(len(pltYColIds)):
                yColId = pltYColIds[i]
                yLabel = self.__procPltData__(data[yColId],pltYDataProcs[i])
                if isNumeric(yLabel):
                    # if the label is a numeric value, take it as the value
                    # and use dataHeader as the yLabel
                    yvalue = yLabel
                    yLabel = str(dataHeader[yColId])
                else:
                    # if the label is not a numerical value, take it as the label
                    # and use 1 as yvalue
                    yvalue = 1
                    yLabel = str(yLabel)
             
                if xLabel in pltData:
                    if yLabel in pltData[xLabel]:
                        pltData[xLabel][yLabel] = pltData[xLabel][yLabel] + yvalue
                    else:
                        pltData[xLabel][yLabel] = yvalue
                else:
                    pltData[xLabel] = {}
                    pltData[xLabel][yLabel] = yvalue

        ##----- 
        ## Mapping the extracted data to the columns for bar chart plot 
        ##----- 

        # the xLabels 
        bar_xlabels = pltData.keys()
        bar_xlabels.sort()

        # collecting values for the bars
        bars = {}
        valueLists = []
        for l in bar_xlabels:
            valueLists.append(pltData[l].keys())

        valueLists = self.__unionLists__(valueLists)

        for v in valueLists:
            barValues = []
            for l in bar_xlabels:
                if v in pltData[l]:
                    barValues.append(pltData[l][v])
                else:
                    barValues.append(0)
            bars[v] = barValues

        # the pltColorMap
        bar_cmap = self.__setColormap__( pltColorMap, len(bars.keys()) )

        ##-----
        ## Generating the bar chart 
        ##-----
        self.output = pltOutput
        self.__setFigureId__()

        num_bars  = len(bars) # the number of bars
        num_dbars = num_bars  # the number of basrs displayed on the plots
        if stackedBar:
            num_dbars  = 1
        num_xtics = len(bar_xlabels)

        figure(self.figId)
        rc('font',size=6.0)
        a = axes([0.1,0.3,0.8,0.6])

        bar_width = 0.35
        bar_gap   = 0.3
        bar_org   = map(lambda x:x*(num_dbars*bar_width+bar_gap), arange(num_xtics))
        tic_pos   = map(lambda x:x+num_dbars*bar_width/2.0, bar_org)

        legend_labels = []
        legend_texts  = bars.keys()

        bar_off = map(lambda x:0,bars[legend_texts[0]])
        for i in range(num_bars):
            chart = None
            if stackedBar and i>0:
                bar_off = map(lambda x,y:x+y,bar_off,bars[legend_texts[i-1]])
                chart   = bar(bar_org, bars[legend_texts[i]], width=bar_width, bottom=bar_off, color=bar_cmap[i])
            else:
                bar_pos = map(lambda x:x+bar_width*i,bar_org)
                chart   = bar(bar_pos, bars[legend_texts[i]], width=bar_width, bottom=bar_off, color=bar_cmap[i])

            legend_labels.append(chart[0])

        grid(True)
        title(pltTitle)
        ylabel(pltYLabel)
        if pltXLabel is None:
            pltXLabel = str(dataHeader[0])
        xlabel(pltXLabel)
        xx = xticks(tic_pos, bar_xlabels)
        setp(xx[1], 'rotation', 90, fontsize=8.0)
        legend(legend_labels, legend_texts, shadow=True, loc='best')
        axis('on')

        self.__doPlot__()

    def __makeMultiHistograms__(self, dataTable, pltColIdList, pltDataProcList, pltLabelList, pltTitle='Histogram',\
                                pltNumBins=50,pltRange=(-1,-1),pltXLabel=None,pltColorMap=None,pltNormalize=False,\
                                pltOutput=None):

        """ backend generator for multiple histograms """

        hist_cmap = self.__setColormap__( pltColorMap, len(pltColIdList) )

        # set figure id
        self.__setFigureId__()

        figure(self.figId)
        rc('font',size=8.0)

        # create only the xlabel if only one histogram is plotted
        if not pltXLabel:
            xlabel(dataHeader[pltColId])
        else:
            xlabel(pltXLabel)

        for id in pltColIdList:

            pltFaceColor = hist_cmap[id]
            pltDataProc  = pltDataProcList[id]
            pltLabel     = pltLabelList[id]

            self.__makeHistogram__(dataTable, pltColId=id, pltNumBins=pltNumBins, pltRange=pltRange, \
                                   pltFaceColor=pltFaceColor, pltNormalize=pltNormalize, \
                                   pltLabel=pltLabel, pltOutput=pltOutput, pltDataProc=pltDataProc)

        # making the plot
        self.output = pltOutput

        if pltNormalize:
            ylabel('')
        else:
            ylabel('# jobs')


        title(pltTitle)
        legend()
        grid(True)
        self.__doPlot__()

    def __makeHistogram__(self,dataTable,pltColId=0,pltNumBins=50, pltRange=(-1,-1), \
                          pltFaceColor='green',pltNormalize=False,\
                          pltLabel=None,pltOutput=None,pltDataProc=None):

        """ backend histogram generator """

        pltAlpha=0.7

        ##-----
        ## extract the given data for pie chart
        ##-----
        # summarize the data of the given column
        dataHeader = dataTable[0]
        dataTable  = dataTable[1:]

        pltData = {}
        pltData[dataHeader[pltColId]] = []
        for data in dataTable:
            
            # apply the user defined data processor
            value = self.__procPltData__(data[pltColId],pltDataProc)

            # make sure the value is in floating point
            if not isNumeric(value):
                raise TypeError('histogram data should be a numeric value.')
            else:
                pltData[dataHeader[pltColId]].append(value)
        
        #a = axes([0.05,0.1,0.5,0.7])
        ## disable plot range if -1 is given to the range specification
        if -1 in pltRange:
            pltRange = None

        n, bins, patches = hist(pltData[dataHeader[pltColId]], bins=pltNumBins, range=pltRange, normed=pltNormalize, facecolor=pltFaceColor, label=pltLabel, alpha=pltAlpha)

        ## a trick for remove the dummy labels from the legend
        for p in patches[1:]:
            p.set_label('_nolegend_')
        

    def __makePieChart__(self,dataTable,pltColId=0,pltTitle='Pie Chart',pltColorMap=None,pltOutput=None,pltDataProc=None):

        """ backend pie chart generator """

        ##----- 
        ## extract the given data for pie chart 
        ##----- 
        # summarize the data of the given column
        dataHeader = dataTable[0]
        dataTable  = dataTable[1:]

        pltData = {}
        for data in dataTable:
            # apply the user defined data processor
            value = self.__procPltData__(data[pltColId],pltDataProc)

            # should make sure all the values are represented in String type
            value = str(value)

            if value in pltData:
                pltData[value] = pltData[value] + 1
            else:
                pltData[value] = 1

        ##----- 
        ## Mapping the extracted data to the columns for pie chart plot 
        ##----- 
        # the pltColorMap
        pie_cmap = self.__setColormap__( pltColorMap, len(pltData.keys()) )
 
        # the labels 
        pie_labels = pltData.keys()
        pie_labels.sort()
 
        # the values
        pie_values = []
        for label in pie_labels:
            pie_values.append(pltData[label])
 
        ##----- 
        ## Generating the pie chart 
        ##----- 
        plotLegend = False
        if len(pie_labels) >= 8: plotLegend = True

        self.output = pltOutput

        self.__setFigureId__()

        if plotLegend:

            legend_length    = 1000 # put a very large number will put all lables in one legend 
            legend_panel_num = len(pie_labels) / legend_length
            if len(pie_labels) % legend_length > 0: legend_panel_num += 1

            fig_size_x  = 6.0 + 4.0 * legend_panel_num
            fig_size_y  = 5.0
            fig_rleft   = 0.05 * 10.0 / fig_size_x 
            fig_rbottom = 0.1
            fig_rwidth  = 0.5  * 10.0 / fig_size_x
            fig_rheight = 0.8

            legend_rwidth = 0.3 * 10.0 / fig_size_x

            figure(self.figId,figsize=(fig_size_x,fig_size_y))
            rc('legend',fontsize=8.0)
            rc('font',size=8.0)
            a = axes([fig_rleft,fig_rbottom,fig_rwidth,fig_rheight])
            chart = pie(pie_values,labels=None,colors=pie_cmap,autopct='%1.1f%%',shadow=True)
            # add percentage on the label
            def sumList(L):
                    return reduce(lambda x,y:x+y, L)
            for i in range(len(pie_labels)):
                percentage = '(%1.1f%%) ' % (100.0 * pie_values[i] / sumList(pie_values))
                pie_labels[i] = percentage + pie_labels[i]

            # put labels on legends
            for i in range(legend_panel_num):
                id_beg = i * legend_length
                id_end = id_beg + legend_length
                if id_end > len(pie_labels): id_end = len(pie_labels)
                #legend_loc = (2*fig_rleft + fig_rwidth + i*(legend_rwidth+fig_rleft), 0.5)
                #legend(chart[0][id_beg:id_end],pie_labels[id_beg:id_end],loc='best',labelsep=0.005,handlelen=0.05,pad=0.1,numpoints=2)
                figlegend(chart[0][id_beg:id_end],pie_labels[id_beg:id_end],loc='center right',labelsep=0.005,handlelen=0.05,pad=0.1,numpoints=2)
        else:
            figure(self.figId)
            rc('font',size=8.0)
            #a = axes([0.05,0.1,0.5,0.7])
            chart = pie(pie_values,labels=pie_labels,colors=pie_cmap,autopct='%1.1f%%',shadow=True)

        title(pltTitle)
        axis('off')
        self.__doPlot__()

    def barchart(self,jobs,xattr,yattr,**keywords):
        
        """
        The plotter's interface for generating bar chart.
        
        usage:
            >>> plotter.barchart(jobs,xattr,yattr,**keywords)

            Generate a bar chart presenting the distribution of yattr along xattr. 

        required arguments:
             jobs: A GANGA's job table

            xattr: A string or a user defined function, the corresponding value of which
                   will be extracted as the x value of the bar chart.

            yattr: A string or a user defined function, the corresponding value of which
                   will be extracted as the y value of the bar chart.

        optional arguments:
            colormap: A list of values representing the colors that will be picked to
                      paint the pie chart. The supported description of color can be found 
                      at http://matplotlib.sourceforge.net/matplotlib.pylab.html#-colors

               title: A string specifying the title of the bar chart.

              xlabel: A string specifying the xlabel of the bar chart.

              ylabel: A string specifying the ylabel of the bar chart.

              output: A name of file where the bar chart plot will be exported. The format
                      is auto-determinated by the extension of the given name.

           xdataproc: An user-defined function which will be applied to process the value 
                      of xattr before plotting the bar chart.

           ydataproc: The same as xdataproc; but apply on the value of yattr. 

            xattrext: Trigger the build-in data pre-processor on the value of xattr.
                      It will be override if "xdataproc" argument is also specified.
                      ** Supported attrext for "backend.actualCE" and "backend.CE":
                         - "by_ce": Catagorize CE queues into CE
                         - "by_country": Catagorize CE queues into country 

            yattrext: The same as xattrext; but apply on the value of yattr.

             stacked: True of False for stacking multiple bars in one stick.

                deep: Sets if looping over all the subjob levels. Default is "True"
        """

        # default keyword arguments
        subtitle  = self.__defaultSubtitle__(yattr)      # default subtitle
        xlabel    = None        # xlabel
        ylabel    = '#'         # ylabel
        colormap  = None        # colormap
        output    = None        # the output file of the picture
        xattrext  = None        # trigger the build-in data processing function on xattr
        yattrext  = None        # trigger the build-in data processing function on yattr
        xdataproc = None        # function for xdata processing (cannot work together with 'attrext')
        ydataproc = None        # function for ydata processing (cannot work together with 'attrext')
        stacked   = True        # stack bars in one stick
        deep      = True        # deep looping over all the subjob levels 

        # update keyword arguments with the given values
        if 'title' in keywords:    subtitle   = keywords['title']
        if 'xlabel' in keywords:   xlabel     = keywords['xlabel']
        if 'ylabel' in keywords:   ylabel     = keywords['ylabel']
        if 'colormap' in keywords: colormap   = keywords['colormap']
        if 'output' in keywords:   output     = keywords['output']
        if 'xattrext' in keywords: xattrext   = keywords['xattrext']
        if 'yattrext' in keywords: yattrext   = keywords['yattrext']
        if 'xdataproc' in keywords: xdataproc = keywords['xdataproc']
        if 'ydataproc' in keywords: ydataproc = keywords['ydataproc']
        if 'stacked' in keywords:   stacked   = keywords['stacked']
        if 'deep' in keywords:       deep     = keywords['deep']

        jlist = []
        for j in jobs:
            jlist.append(j)

        xattr_spec = self.__makeList__(xattr)
        yattr_spec = self.__makeList__(yattr)

        # special build-in dataprocs for the CE-based bar chart
        xdataproc = self.__setDataProcessor__(xattr_spec[0],xattrext,xdataproc)
        ydataproc = self.__setDataProcessor__(yattr_spec[0],yattrext,ydataproc)

        dataTable = getJobInfoTable(jlist,[xattr_spec[0],yattr_spec[0]],deep)
        # make the plot title
        title = __makePlotTitle__(len(dataTable[1:]),deep,subtitle)

        # make the plot
        self.__makeBarChart__(dataTable,pltXColId=0,pltYColIds=[1],pltTitle=title,pltXLabel=xlabel,pltYLabel=ylabel,pltColorMap=colormap,pltOutput=output,pltXDataProc=xdataproc,pltYDataProcs=[ydataproc],stackedBar=stacked)

    def scatter(self,jobs,xattr,yattr,**keywords):

        """
        The plotter's interface for generating scatter plot.

        usage:
            >>> plotter.scatter(jobs,xattr,yattr,**keywords)

            Generate a scatter chart presenting the distribution of yattr along xattr.

        required arguments:
             jobs: A GANGA's job table

            xattr: A string or a user defined function, the corresponding value of which
                   will be extracted as the x value of the scatter plot.

            yattr: A string or a user defined function, the corresponding value of which
                   will be extracted as the y value of the scatter plot.

        optional arguments:

               cattr: A String or a user defined function to catagorize the data points into group.
                      Each group will be plotted with the same marker filled with the same color.

            colormap: A list of values representing the colors that will be picked to
                      distinguish different data group categorized by cattr. The supported description of color can be found
                      at http://matplotlib.sourceforge.net/matplotlib.pylab.html#-colors

               title: A string specifying the title.

              xlabel: A string specifying the xlabel.

              ylabel: A string specifying the ylabel.

              output: A name of file where the chart will be exported. The format
                      is auto-determinated by the extension of the given name.

           xdataproc: An user-defined function which will be applied to process the value
                      of xattr before plotting chart.

           ydataproc: The same as xdataproc; but apply on the value of yattr.

           cdataproc: The same as xdataproc; but apply on the value of cattr.

            xattrext: Trigger the build-in data pre-processor on the value of xattr.
                      It will be override if "xdataproc" argument is also specified.
                      ** Supported attrext for "backend.actualCE" and "backend.CE":
                         - "by_ce": Catagorize CE queues into CE
                         - "by_country": Catagorize CE queues into country

            yattrext: The same as xattrext; but apply on the value of yattr.

            cattrext: The same as xattrext; but apply on the value of cattr.

                deep: Sets if looping over all the subjob levels. Default is "True"
        """

        # default keyword arguments
        subtitle  = self.__defaultSubtitle__(yattr)      # default subtitle
        cattr     = None        # catagory attribute
        xlabel    = None        # xlabel
        ylabel    = None        # ylabel
        colormap  = None        # colormap
        output    = None        # the output file of the picture
        xattrext  = None        # trigger the build-in data processing function on xattr
        yattrext  = None        # trigger the build-in data processing function on yattr
        cattrext  = None        # trigger the build-in data processing function on cattr
        xdataproc = None        # function for xdata processing (cannot work together with 'attrext')
        ydataproc = None        # function for ydata processing (cannot work together with 'attrext')
        cdataproc = None        # function for cdata processing (cannot work together with 'attrext')
        deep      = True        # deep looping over all the subjob levels

        # update keyword arguments with the given values
        if 'cattr' in keywords:     cattr     = keywords['cattr']
        if 'title' in keywords:    subtitle   = keywords['title']
        if 'xlabel' in keywords:   xlabel     = keywords['xlabel']
        if 'ylabel' in keywords:   ylabel     = keywords['ylabel']
        if 'colormap' in keywords: colormap   = keywords['colormap']
        if 'output' in keywords:   output     = keywords['output']
        if 'xattrext' in keywords: xattrext   = keywords['xattrext']
        if 'yattrext' in keywords: yattrext   = keywords['yattrext']
        if 'cattrext' in keywords: cattrext   = keywords['cattrext']
        if 'xdataproc' in keywords: xdataproc = keywords['xdataproc']
        if 'ydataproc' in keywords: ydataproc = keywords['ydataproc']
        if 'cdataproc' in keywords: cdataproc = keywords['cdataproc']
        if 'deep' in keywords:       deep     = keywords['deep']

        jlist = []
        for j in jobs:
            jlist.append(j)

        xattr_spec = self.__makeList__(xattr)
        yattr_spec = self.__makeList__(yattr)
        cattr_spec = self.__makeList__(cattr)

        # special build-in dataprocs for the CE-based bar chart
        xdataproc = self.__setDataProcessor__(xattr_spec[0],xattrext,xdataproc)
        ydataproc = self.__setDataProcessor__(yattr_spec[0],yattrext,ydataproc)
        cdataproc = self.__setDataProcessor__(cattr_spec[0],cattrext,cdataproc)

        if cattr_spec[0]:
            dataTable = getJobInfoTable(jlist,[xattr_spec[0],yattr_spec[0],cattr_spec[0]],deep)
        else:
            dataTable = getJobInfoTable(jlist,[xattr_spec[0],yattr_spec[0]],deep)
            
        # make the plot title
        title = __makePlotTitle__(len(dataTable[1:]),deep,subtitle)

        # make the plot
        self.__makeScatter__(dataTable,pltXColId=0,pltYColId=1,pltTitle=title,pltXLabel=xlabel,pltYLabel=ylabel,pltColorMap=colormap,pltOutput=output,pltXDataProc=xdataproc,pltYDataProc=ydataproc,pltCDataProc=cdataproc)

    def piechart(self,jobs,attr,**keywords):
        
        """
        The plotter's interface for generating pie chart.
        
        usage:
            >>> plotter.piechart(jobs,attr,**keywords)

            Generate a pie chart presenting the distribution of th attr among
            the given jobs.

        required arguments:
            jobs: A GANGA's job table

            attr: A string or a user defined function, the corresponding value of which
                  will be extracted from the given jobs to generate the pie chart.

        optional arguments:
            colormap: A list of values representing the colors that will be picked to
                      paint the pie chart. The supported description of color can be found 
                      at http://matplotlib.sourceforge.net/matplotlib.pylab.html#-colors

               title: A string specifying the title of the pie chart.

              output: A name of file where the pie chart plot will be exported. The format
                      is auto-determinated by the extension of the given name.

            dataproc: An user-defined function which will be applied to process the value 
                      of attr before plotting the pie chart.

             attrext: Trigger the build-in data pre-processor on the value of attr.
                      It will be override if "dataproc" argument is also specified.
                      ** Supported attrext for "backend.actualCE" and "backend.CE":
                         - "by_ce": Catagorize CE queues into CE
                         - "by_country": Catagorize CE queues into country 

                deep: Sets if looping over all the subjob levels. Default is "True"
        """

        # default keyword arguments
        subtitle = self.__defaultSubtitle__(attr)       # default subtitle
        colormap = None        # colormap
        output   = None        # the output file of the picture
        attrext  = None        # trigger the build-in data processing function 
        dataproc = None        # function for data processing (cannot work together with 'attrext')
        deep     = True        # deep looping over all the subjob levels 

        # update keyword arguments with the given values
        if 'title' in keywords:    subtitle = keywords['title']
        if 'colormap' in keywords: colormap = keywords['colormap']
        if 'output' in keywords:   output   = keywords['output']
        if 'attrext' in keywords:  attrext  = keywords['attrext']
        if 'dataproc' in keywords: dataproc = keywords['dataproc']
        if 'deep' in keywords:       deep   = keywords['deep']

        jlist = []
        for j in jobs:
            jlist.append(j)

        attr_spec = self.__makeList__(attr)

        # special build-in dataprocs for the CE-based pie chart
        dataproc  = self.__setDataProcessor__(attr,attrext,dataproc)
        dataTable = getJobInfoTable(jlist,attr_spec,deep)

        # make the plot title
        title = __makePlotTitle__(len(dataTable[1:]),deep,subtitle)

        # make the plot
        self.__makePieChart__(dataTable,pltColId=0,pltTitle=title,pltColorMap=colormap,pltOutput=output,pltDataProc=dataproc)

    def histogram(self,jobs,attr,**keywords):

        """
        The plotter's interface for generating one or multiple histograms in one chart.

        usage:
            >>> plotter.histogram(jobs,attr,**keywords)

            Generate a histogram catogorized by the attr among the given jobs.

        required arguments:
            jobs: A GANGA's job table

            attr: A string or a user defined function, the corresponding value of which
                  will be extracted from the given jobs to generate the histogram.

        optional arguments:
            colormap: A list of values representing the colors that will be picked to
                      paint the histogram bars. The supported description of color can be found
                      at http://matplotlib.sourceforge.net/matplotlib.pylab.html#-colors

           normalize: Sets if doing histogram normalization. Default is "False"

             numbins: Sets the number of bins in the histogram. Default is 50

               title: A string specifying the title of the histogram.

              xlabel: A string specifying the xlabel of the histogram.

               label: A string specifying the lable of the histogram displayed on the legend.

                xmin: A number specifying the lower bound of the histogram range.
    
                xmax: A number specifying the upper bound of the histogram range.

              output: A name of file where the pie chart plot will be exported. The format
                      is auto-determinated by the extension of the given name.

            dataproc: An user-defined function which will be applied to process the value
                      of attr before plotting the histogram.

                deep: Sets if looping over all the subjob levels. Default is "True"
        """

        # default keyword arguments
        subtitle  = self.__defaultSubtitle__(attr)       # default subtitle
        xlabel    = self.__defaultSubtitle__(attr)       # default xlabel
        label     = self.__defaultSubtitle__(attr)       # default label
        colormap  = None        # colormap
        output    = None        # the output file of the picture
        attrext   = None        # trigger the build-in data processing function
        dataproc  = None        # function for data processing (cannot work together with 'attrext')
        deep      = True        # deep looping over all the subjob levels
        normalize = False       # histogram normalization
        numbins   = 50          # number of bins in the histogram
        xmin      = -1          # lower bound of the histogram 
        xmax      = -1          # upper bound of the histogram

        # update keyword arguments with the given values
        if 'title' in keywords:    subtitle  = keywords['title']
        if 'colormap' in keywords: colormap  = keywords['colormap']
        if 'xlabel' in keywords:    xlabel   = keywords['xlabel']
        if 'label' in keywords:     label    = keywords['label']
        if 'output' in keywords:   output    = keywords['output']
        if 'dataproc' in keywords: dataproc  = keywords['dataproc']
        if 'deep' in keywords:       deep    = keywords['deep']
        if 'normalize' in keywords: normalize= keywords['normalize']
        if 'numbins' in keywords:   numbins  = keywords['numbins']
        if 'xmin' in keywords:      xmin     = keywords['xmin']
        if 'xmax' in keywords:      xmax     = keywords['xmax']

        jlist = []
        for j in jobs:
            jlist.append(j)

        attr_spec = self.__makeList__(attr)

        logger.debug('attribute specification: %s' % repr(attr_spec))

        # special build-in dataprocs for the CE-based pie chart
        dataproc  = self.__setDataProcessor__(attr,attrext,dataproc)
        dataTable = getJobInfoTable(jlist,attr_spec,deep)

        logger.debug('internal data table: %s' % repr(dataTable))

        # make the plot title
        title = __makePlotTitle__(len(dataTable[1:]),deep,subtitle)

        pltColIdList = range(len(dataTable[0]))

        if not canLoopOver(dataproc):
            dataproc = len( dataTable[0] ) * [dataproc]
        elif len(dataproc) != len(dataTable[0]):
            logger.error('dataproc requires a list with %d elements, %d are given' % (len(dataTable[0]),len(dataproc)) )
            return False

        if (not canLoopOver(label)) or (isStringLike(label)):
            label = len( dataTable[0] ) * [label]
        elif len(label) != len(dataTable[0]):
            logger.error('label requires a list with %d elements, %d are given' % (len(dataTable[0]), len(label)) )
            return False

        # make the plot
        self.__makeMultiHistograms__(dataTable,pltColIdList=pltColIdList,pltDataProcList=dataproc,pltLabelList=label, \
                                     pltTitle=title,pltNumBins=numbins,pltRange=(xmin,xmax), pltXLabel=xlabel, \
                                     pltColorMap=colormap,pltNormalize=normalize, \
                                     pltOutput=output)




#!/usr/bin/env python

import os, re, time, commands, socket
from dq2.tracer.client.TracerClient import TracerClient
from dq2.common import generate_uuid

try:
    import hashlib
    md = hashlib.md5()
except ImportError:
    # for Python << 2.5
    import md5
    md = md5.new()

########################################################################
def fill_report(uuid = None, eventVersion = None, remoteSite=None,
                localSite=None, timeStart=None, catStart=None,
                relativeStart=None, transferStart=None, validateStart=None,
                timeEnd=None, duid=None,version=None, dataset=None,
                protocol=None, filename=None, filesize=None,
                guid=None, usr=None, 
                hostname=None, ip=None, suspicious=0, appid=None, usrdn=None  ):
    if uuid:
        uuid = uuid.replace('-','')
    if duid:
        duid = duid.replace('-','')
    if guid:
        guid = guid.replace('-','')
    if usr:
        usr = usr.replace('-','')
    report = {
        'uuid': uuid,
        'eventType': 'drct_ganga',
        'eventVersion': eventVersion,
        'remoteSite': remoteSite,
        'localSite': localSite,
        'timeStart': timeStart,
        'catStart': catStart,
        'relativeStart': relativeStart,
        'transferStart': transferStart,
        'validateStart': validateStart,
        'timeEnd': timeEnd,
        'duid': duid,
        'version': version,
        'dataset': dataset,
        'clientState': 'INIT_REPORT',
        'protocol': protocol,
        'filename': filename,
        'filesize': filesize,
        'guid': guid,
        'usr': usr, 
        'hostname': hostname,
        'ip': ip,
        'suspicious': suspicious, 
        'appid': appid,
        'usrdn': usrdn 
    }
    
    return report

########################################################################
if __name__ == '__main__':

    starttime = time.time()
    print '--------------'
    print 'DQ2 tracer preparation ...'

    usrDN = commands.getstatusoutput('voms-proxy-info -identity')[1]
    # Event uuid
    uuid = generate_uuid()
    # Read LFC and transfer time produced in ganga-stage-in-out-dq2.py
    # or make-filestager-joboption.py
    tracertimes = [ line.strip() for line in file('dq2tracertimes.txt') ]
    # User hash
    #m=md5.new()
    md.update(usrDN)
    usrhex = md.hexdigest()

    numfiles3 = 0
    filepat = '"PFN:(.+)"'
    selectorpat = 'ServiceMgr.EventSelector.InputCollections = \[(.+),\]'
    selectorpat2 = 'mySampleList = \[(.+)\]'
    sitepat = 'detected DQ2_LOCAL_SITE_ID: (.+)'
    datasetnamepat = '^DATASETNAME=(.+)'
    
    input_files = []
    input_files_sel = []

    if 'stdout' in os.listdir('.'):  
        zfile = os.popen('cat '+os.path.join('.','stdout' ))
        for line in zfile:
            # Find EventSelector
            match = re.search(selectorpat, line)
            if match:
                selfiles = match.group(1)
                selfiles = re.sub('"','', selfiles)
                input_files_sel = selfiles.split(',')
                
            # Find mySampleList
            match = re.search(selectorpat2, line)
            if match:
                selfiles = match.group(1)
                selfiles = re.sub('"','', selfiles)
                input_files_sel = selfiles.split(',')
            
            # Find actually processed input files
            match = re.search(filepat, line)
            if match:
                filename = match.group(1)
                if filename and not filename in input_files:
                    input_files.append(filename)
                    numfiles3 = numfiles3 + 1
                    
            match = re.search(sitepat, line)       
            if match:
                siteid = match.group(1)
                print siteid
                
            match = re.search(datasetnamepat, line)       
            if match:
                datasetname = match.group(1)

        if zfile:        
            zfile.close()

    # Read in info from athena summary AthSummary.txt and use this info 
    athinfo = {}
    if 'AthSummary.txt' in os.listdir('.'):  
        readline = False
        zfile = os.popen('cat '+os.path.join('.','AthSummary.txt' ))
        for line in zfile:
            if line.find('Files read:')==0:
                numfiles3 = int(re.match('Files read:(.*)',line).group(1))
            if line.find('Exit Status:')==0:
                readline=True
                continue
            try:
                if readline:
                    athinfo = eval(line)
                    readline = False
            except:
                pass

        if zfile:        
            zfile.close()

        try:
            dum = athinfo['files']['read']
            if dum:
                input_files = dum.split(',')
        except:
            pass

    print '%s out of %s file of dataset %s have been processed.' %(numfiles3, len(input_files_sel), datasetname)
    print 'Number could be wrong for certain input modes!'
    
    if 'GANGA_VERSION' in os.environ:
        ganga_version = os.environ['GANGA_VERSION']
    else:
        ganga_version = 'GANGA-5'

    try:
        lfns = [ line.strip() for line in file('input_files') ]
        guids = [ line.strip() for line in file('input_guids') ]
        input_files_txt = [ line.strip() for line in file('input.txt') ]
    except:
        lfns = []
        guids = []
        input_files_txt = []
                
    ddmFileMap = {}
    for i in xrange(0,len(lfns)):
        ddmFileMap[lfns[i]] = guids[i]

    #from dq2.clientapi.DQ2 import DQ2
    #dq2=DQ2()
    #duid = dq2.listDatasets(datasetname)[datasetname]['duid']

    protpat = re.compile(r'^(\w+):/')
    filenamepat = re.compile(r'.*/(.+)$')    

    # Fail over for empty input_files
    if not input_files:
        input_files = input_files_txt

    # Determine new variables
    hostName = socket.gethostbyaddr(socket.gethostname())[0]
    ipAddr = socket.gethostbyaddr(socket.gethostname())[2][0]

    nreport = 0

    for longfilename in input_files:
        prot = re.findall(protpat, longfilename)
        if prot:
            prot = prot[0]
        else:
            prot = 'gridcopy'
        
        lfn = re.findall(filenamepat, longfilename)
        if lfn:
            lfn = lfn[0]
            lfn = re.sub('^tcf_','',lfn)
            lfn = re.sub('(__DQ2.*)$','',lfn)
        else:
            lfn = longfilename

        report = fill_report(uuid = uuid,
                             eventVersion  = ganga_version,
                             remoteSite    = siteid,
                             localSite     = siteid,
                             timeStart     = tracertimes[0],
                             catStart      = tracertimes[1],
                             relativeStart = tracertimes[2],
                             transferStart = tracertimes[3],
                             timeEnd       = starttime,
                             dataset       = datasetname,
                             protocol      = prot,
                             filename      = lfn,
                             guid = ddmFileMap[lfn],
                             #duid = duid,
                             #filesize = contents[ddmFileMap[lfn]]['filesize'],
                             duid = None,
                             filesize = -1,
                             version = 0,
                             usr = usrhex,
                             hostname = hostName,
                             ip = ipAddr,
                             suspicious = 0, 
                             appid = None,
                             usrdn = usrDN
                             )
        #print report
        try:
            #TracerClient().addReport(report)
            nreport+=1
        except:
            print 'An Error during TracerClient().addReport(report) occured'
            print report
            pass

    print '%s DQ2 tracer file reports have been sent.' %nreport
    print '--------------'




###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id$
###############################################################################
# Athena Jedi Runtime Handler
#
# ATLAS/ARDA

import os, sys, pwd, commands, re, shutil, urllib, time, string, exceptions, random, fnmatch

from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

from GangaAtlas.Lib.ATLASDataset import DQ2Dataset, DQ2OutputDataset
from GangaPanda.Lib.Panda.Panda import runPandaBrokerage, uploadSources, getLibFileSpecFromLibDS
from GangaCore.Core.exceptions import BackendError

from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2outputdatasetname
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_set_dataset_lifetime
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_lock, dq2

from GangaCore.Utility.GridShell import getShell

def createContainer(name):
    from pandatools import Client
    # don't create containers for HC datasets
    if not configPanda['processingType'].startswith('gangarobot') and not configPanda['processingType'].startswith('hammercloud'):
        try:
            Client.createContainer(name,False)
            logger.info('Created output container %s' %name)
        except exceptions.SystemExit:
            raise BackendError('Panda','Exception in Client.createContainer %s: %s %s'%(name,sys.exc_info()[0],sys.exc_info()[1]))

def addDatasetsToContainer(container,datasets):
    from pandatools import Client
    # HC datasets don't use containers
    if not configPanda['processingType'].startswith('gangarobot') and not configPanda['processingType'].startswith('hammercloud'):
        Client.addDatasetsToContainer(container,datasets,False)


def getDBDatasets(jobO,trf,dbrelease):
    from pandatools import Client

    # get DB datasets
    dbrFiles  = {}
    dbrDsList = []
    if trf or dbrelease != '':
        if trf:
            # parse jobO for TRF
            tmpItems = jobO.split()
        else:
            # mimic a trf parameter to reuse following algorithm
            tmpItems = ['%DB='+dbrelease]
        # look for DBRelease
        for tmpItem in tmpItems:
            match = re.search('%DB=([^:]+):(.+)$',tmpItem)
            if match:
                tmpDbrDS  = match.group(1)
                tmpDbrLFN = match.group(2)
                # get files in the dataset
                if not tmpDbrDS in dbrDsList:
                    logger.info("Querying files in dataset:%s" % tmpDbrDS)
                    try:
                        tmpList = Client.queryFilesInDataset(tmpDbrDS,False)
                    except:
                        raise ApplicationConfigurationError("ERROR : error while looking up dataset %s. Perhaps this dataset does not exist?"%tmpDbrDS)
                    # append
                    for tmpLFN,tmpVal in tmpList.iteritems():
                        dbrFiles[tmpLFN] = tmpVal
                    dbrDsList.append(tmpDbrDS)
                # check
                if tmpDbrLFN not in dbrFiles:
                    raise ApplicationConfigurationError("ERROR : %s is not in %s"%(tmpDbrLFN,tmpDbrDS))
    return dbrFiles,dbrDsList

def expandExcludedSiteList( job ):
    '''Expand a site list taking wildcards into account'''
                
    # first, check if there's anything to be done
    check_ddm = False
    wildcard = False
    excl_sites = []
    for s in job.backend.requirements.excluded_sites:
        if s.find("ANALY_") == -1:
            check_ddm = True

        if s.find("*") != -1:
            wildcard = True
            
        if s.find("ANALY_") != -1 and s.find("*") == -1:
            excl_sites.append(s)

    if not check_ddm and not wildcard:
        return excl_sites

    # we have either wildcards or DDM sites listed
    # First, find the allowed sites for this job and ensure no duplicates anywhere
    from pandatools import Client
    logger.info("Excluding DDM and wildcarded sites from Jedi job. Please wait....")
    orig_ddm_list = []
    new_ddm_list = []
    for s in job.inputdata.get_locations():
        if not s in orig_ddm_list:
            orig_ddm_list.append(s)
            new_ddm_list.append(s)

    orig_panda_list = []
    for s in [Client.convertDQ2toPandaID(x) for x in new_ddm_list]:        
        for s2 in Client.PandaSites.keys():
            if s2.find(s) != -1 and not s2 in orig_panda_list:
                orig_panda_list.append(s2)

    if check_ddm:
        # remove any DDM sites that are referenced, including wildcards
        for s in job.backend.requirements.excluded_sites:
            if s in orig_ddm_list:
                new_ddm_list.remove(s)

            if s.find("*") != -1:
                for s2 in orig_ddm_list:
                    if fnmatch.fnmatch(s2, s):
                        new_ddm_list.remove(s2)
                        
        # now recreate the panda list and see if any have been dropped
        new_panda_list = []
        for s in [Client.convertDQ2toPandaID(x) for x in new_ddm_list]:        
            for s2 in Client.PandaSites.keys():
                if s2.find(s) != -1 and not s2 in new_panda_list:
                    new_panda_list.append(s)

        for s in orig_panda_list:
            if not s in new_panda_list and not s in excl_sites:
                excl_sites.append(s)
                
    if wildcard:
        # find wilcarded ANALY_* sites and exclude any that match good sites
        for s in job.backend.requirements.excluded_sites:
            if s.find("*") != -1:
                for s2 in orig_panda_list:
                    if fnmatch.fnmatch(s2, s) and not s2 in excl_sites:
                        excl_sites.append(s2)
                        
    return excl_sites

class AthenaJediRTHandler(IRuntimeHandler):
    '''Athena Jedi Runtime Handler'''


    def master_prepare(self,app,appconfig):
        '''Prepare the master job'''

        from pandatools import Client
        from pandatools import MiscUtils
        from pandatools import AthenaUtils
        from pandatools import PsubUtils

        # create a random number for this submission to allow multiple use of containers
        self.rndSubNum = random.randint(1111,9999)

        job = app._getParent()
        logger.debug('AthenaJediRTHandler master_prepare called for %s', job.getFQID('.')) 

        if app.useRootCoreNoBuild:
            logger.info('Athena.useRootCoreNoBuild is True, setting Panda.nobuild=True.')
            job.backend.nobuild = True

        if job.backend.bexec and job.backend.nobuild:
            raise ApplicationConfigurationError("Contradicting options: job.backend.bexec and job.backend.nobuild are both enabled.")

        if job.backend.requirements.rootver != '' and job.backend.nobuild:
            raise ApplicationConfigurationError("Contradicting options: job.backend.requirements.rootver given and job.backend.nobuild are enabled.")
        
        # Switch on compilation flag if bexec is set or libds is empty
        if job.backend.bexec != '' or not job.backend.nobuild:
            app.athena_compile = True
            for sj in job.subjobs:
                sj.application.athena_compile = True
            logger.info('"job.backend.nobuild=False" or "job.backend.bexec" is set - Panda build job is enabled.')

        if job.backend.nobuild:
            app.athena_compile = False
            for sj in job.subjobs:
                sj.application.athena_compile = False
            logger.info('"job.backend.nobuild=True" or "--nobuild" chosen - Panda build job is switched off.')

        # check for auto datri
        if job.outputdata.location != '':
            if not PsubUtils.checkDestSE(job.outputdata.location,job.outputdata.datasetname,False):
                raise ApplicationConfigurationError("Problems with outputdata.location setting '%s'" % job.outputdata.location)

        # validate application
        if not app.atlas_release and not job.backend.requirements.rootver and not app.atlas_exetype in [ 'EXE' ]:
            raise ApplicationConfigurationError("application.atlas_release is not set. Did you run application.prepare()")

        self.dbrelease = app.atlas_dbrelease
        if self.dbrelease != '' and self.dbrelease != 'LATEST' and self.dbrelease.find(':') == -1:
            raise ApplicationConfigurationError("ERROR : invalid argument for DB Release. Must be 'LATEST' or 'DatasetName:FileName'")

        self.runConfig = AthenaUtils.ConfigAttr(app.atlas_run_config)
        for k in self.runConfig.keys():
            self.runConfig[k]=AthenaUtils.ConfigAttr(self.runConfig[k])
        if not app.atlas_run_dir:
            raise ApplicationConfigurationError("application.atlas_run_dir is not set. Did you run application.prepare()")
 
        self.rundirectory = app.atlas_run_dir
        self.cacheVer = ''
        if app.atlas_project and app.atlas_production:
            self.cacheVer = "-" + app.atlas_project + "_" + app.atlas_production

        # handle different atlas_exetypes
        self.job_options = ''
        if app.atlas_exetype == 'TRF':
            self.job_options += ' '.join([os.path.basename(fopt.name) for fopt in app.option_file])

            #if not job.outputdata.outputdata:
            #    raise ApplicationConfigurationError("job.outputdata.outputdata is required for atlas_exetype in ['PYARA','ARES','TRF','ROOT','EXE' ] and Panda backend")
            #raise ApplicationConfigurationError("Sorry TRF on Panda backend not yet supported")

            if app.options:
                self.job_options += ' %s ' % app.options
                
        elif app.atlas_exetype == 'ATHENA':
            
            if len(app.atlas_environment) > 0 and app.atlas_environment[0].find('DBRELEASE_OVERRIDE')==-1:
                logger.warning("Passing of environment variables to Athena using Panda not supported. Ignoring atlas_environment setting.")
                
            if job.outputdata.outputdata:
                raise ApplicationConfigurationError("job.outputdata.outputdata must be empty if atlas_exetype='ATHENA' and Panda backend is used (outputs are auto-detected)")
            if app.options:
                if app.options.startswith('-c'):
                    self.job_options += ' %s ' % app.options
                else:
                    self.job_options += ' -c %s ' % app.options

                logger.warning('The value of j.application.options has been prepended with " -c " ')
                logger.warning('Please make sure to use proper quotes for the values of j.application.options !')

            self.job_options += ' '.join([os.path.basename(fopt.name) for fopt in app.option_file])

            # check for TAG compression
            if 'subcoll.tar.gz' in app.append_to_user_area:
                self.job_options = ' uncompress.py ' + self.job_options
                
        elif app.atlas_exetype in ['PYARA','ARES','ROOT','EXE']:

            #if not job.outputdata.outputdata:
            #    raise ApplicationConfigurationError("job.outputdata.outputdata is required for atlas_exetype in ['PYARA','ARES','TRF','ROOT','EXE' ] and Panda backend")
            self.job_options += ' '.join([os.path.basename(fopt.name) for fopt in app.option_file])

            # sort out environment variables
            env_str = ""
            if len(app.atlas_environment) > 0:
                for env_var in app.atlas_environment:
                    env_str += "export %s ; " % env_var
            else: 
                env_str = ""

            # below fixes issue with runGen -- job_options are executed by os.system when dbrelease is used, and by the shell otherwise
            ## - REMOVED FIX DUE TO CHANGE IN PILOT - MWS 8/11/11
            if job.backend.requirements.usecommainputtxt:
                input_str = '/bin/echo %IN > input.txt; cat input.txt; '
            else:
                input_str = '/bin/echo %IN | sed \'s/,/\\\n/g\' > input.txt; cat input.txt; '
            if app.atlas_exetype == 'PYARA':
                self.job_options = env_str + input_str + ' python ' + self.job_options
            elif app.atlas_exetype == 'ARES':
                self.job_options = env_str + input_str + ' athena.py ' + self.job_options
            elif app.atlas_exetype == 'ROOT':
                self.job_options = env_str + input_str + ' root -b -q ' + self.job_options
            elif app.atlas_exetype == 'EXE':
                self.job_options = env_str + input_str + self.job_options

            if app.options:
                self.job_options += ' %s ' % app.options

        if self.job_options == '':
            raise ApplicationConfigurationError("No Job Options found!")
        logger.info('Running job options: %s'%self.job_options)

        # validate dbrelease
        if self.dbrelease != "LATEST":
            self.dbrFiles,self.dbrDsList = getDBDatasets(self.job_options,'',self.dbrelease)

        # handle the output dataset
        if job.outputdata:
            if job.outputdata._name != 'DQ2OutputDataset':
                raise ApplicationConfigurationError('Panda backend supports only DQ2OutputDataset')
        else:
            logger.info('Adding missing DQ2OutputDataset')
            job.outputdata = DQ2OutputDataset()

        # validate the output dataset name (and make it a container)
        job.outputdata.datasetname,outlfn = dq2outputdatasetname(job.outputdata.datasetname, job.id, job.outputdata.isGroupDS, job.outputdata.groupname)
        if not job.outputdata.datasetname.endswith('/'):
            job.outputdata.datasetname+='/'

        # add extOutFiles
        self.extOutFile = []
        for tmpName in job.outputdata.outputdata:
            if tmpName != '':
                self.extOutFile.append(tmpName)
        for tmpName in job.backend.extOutFile:
            if tmpName != '':
                self.extOutFile.append(tmpName)

        # use the shared area if possible
        tmp_user_area_name = app.user_area.name
        if app.is_prepared is not True:
            from GangaCore.Utility.files import expandfilename
            shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])
            tmp_user_area_name = os.path.join(os.path.join(shared_path,app.is_prepared.name),os.path.basename(app.user_area.name))


        # Add inputsandbox to user_area
        if job.inputsandbox:
            logger.warning("Submitting Panda job with inputsandbox. This may slow the submission slightly.")

            if tmp_user_area_name:
                inpw = os.path.dirname(tmp_user_area_name)
                self.inputsandbox = os.path.join(inpw, 'sources.%s.tar' % commands.getoutput('uuidgen 2> /dev/null'))
            else:
                inpw = job.getInputWorkspace()
                self.inputsandbox = inpw.getPath('sources.%s.tar' % commands.getoutput('uuidgen 2> /dev/null'))

            if tmp_user_area_name:
                rc, output = commands.getstatusoutput('cp %s %s.gz' % (tmp_user_area_name, self.inputsandbox))
                if rc:
                    logger.error('Copying user_area failed with status %d',rc)
                    logger.error(output)
                    raise ApplicationConfigurationError('Packing inputsandbox failed.')
                rc, output = commands.getstatusoutput('gunzip %s.gz' % (self.inputsandbox))
                if rc:
                    logger.error('Unzipping user_area failed with status %d',rc)
                    logger.error(output)
                    raise ApplicationConfigurationError('Packing inputsandbox failed.')

            for fname in [os.path.abspath(f.name) for f in job.inputsandbox]:
                fname.rstrip(os.sep)
                path = os.path.dirname(fname)
                fn = os.path.basename(fname)

                #app.atlas_run_dir
                # get Athena versions
                rc, out = AthenaUtils.getAthenaVer()
                # failed
                if not rc:
                    #raise ApplicationConfigurationError('CMT could not parse correct environment ! \n Did you start/setup ganga in the run/ or cmt/ subdirectory of your athena analysis package ?')
                    logger.warning("CMT could not parse correct environment for inputsandbox - will use the atlas_run_dir as default")
                    
                    # as we don't have to be in the run dir now, create a copy of the run_dir directory structure and use that
                    input_dir = os.path.dirname(self.inputsandbox)
                    run_path = "%s/sbx_tree/%s" % (input_dir, app.atlas_run_dir)
                    rc, output = commands.getstatusoutput("mkdir -p %s" % run_path)
                    if not rc:
                        # copy this sandbox file
                        rc, output = commands.getstatusoutput("cp %s %s" % (fname, run_path))
                        if not rc:
                            path = os.path.join(input_dir, 'sbx_tree')
                            fn = os.path.join(app.atlas_run_dir, fn)
                        else:
                            raise ApplicationConfigurationError("Couldn't copy file %s to recreate run_dir for input sandbox" % fname)
                    else:
                        raise ApplicationConfigurationError("Couldn't create directory structure to match run_dir %s for input sandbox" % run_path)

                else:
                    userarea = out['workArea']

                    # strip the path from the filename if present in the userarea
                    ua = os.path.abspath(userarea)
                    if ua in path:
                        fn = fname[len(ua)+1:]
                        path = ua

                rc, output = commands.getstatusoutput('tar -h -r -f %s -C %s %s' % (self.inputsandbox, path, fn))
                if rc:
                    logger.error('Packing inputsandbox failed with status %d',rc)
                    logger.error(output)
                    raise ApplicationConfigurationError('Packing inputsandbox failed.')

            # remove sandbox tree if created
            if "sbx_tree" in os.listdir(os.path.dirname(self.inputsandbox)):                
                rc, output = commands.getstatusoutput("rm -r %s/sbx_tree" % os.path.dirname(self.inputsandbox))
                if rc:
                    raise ApplicationConfigurationError("Couldn't remove directory structure used for input sandbox")
                
            rc, output = commands.getstatusoutput('gzip %s' % (self.inputsandbox))
            if rc:
                logger.error('Packing inputsandbox failed with status %d',rc)
                logger.error(output)
                raise ApplicationConfigurationError('Packing inputsandbox failed.')
            self.inputsandbox += ".gz"
        else:
            self.inputsandbox = tmp_user_area_name

        # job name
        jobName = 'ganga.%s' % MiscUtils.wrappedUuidGen()

        # make task
        taskParamMap = {}
        # Enforce that outputdataset name ends with / for container
        if not job.outputdata.datasetname.endswith('/'):
            job.outputdata.datasetname = job.outputdata.datasetname + '/'

        taskParamMap['taskName'] = job.outputdata.datasetname

        taskParamMap['uniqueTaskName'] = True
        taskParamMap['vo'] = 'atlas'
        taskParamMap['architecture'] = AthenaUtils.getCmtConfig(athenaVer=app.atlas_release, cmtConfig=app.atlas_cmtconfig)
        if app.atlas_release:
            taskParamMap['transUses'] = 'Atlas-%s' % app.atlas_release
        else:
            taskParamMap['transUses'] = ''
        taskParamMap['transHome'] = 'AnalysisTransforms'+self.cacheVer#+nightVer

        configSys = getConfig('System')
        gangaver = configSys['GANGA_VERSION'].lower()
        if not gangaver:
            gangaver = "ganga"

        if app.atlas_exetype in ["ATHENA", "TRF"]:
            taskParamMap['processingType'] = '{0}-jedi-athena'.format(gangaver)
        else:
            taskParamMap['processingType'] = '{0}-jedi-run'.format(gangaver)

        #if options.eventPickEvtList != '':
        #    taskParamMap['processingType'] += '-evp'
        taskParamMap['prodSourceLabel'] = 'user'
        if job.backend.site != 'AUTO':
            taskParamMap['cloud'] = Client.PandaSites[job.backend.site]['cloud']
            taskParamMap['site'] = job.backend.site
        elif job.backend.requirements.cloud is not None and not job.backend.requirements.anyCloud:
            taskParamMap['cloud'] = job.backend.requirements.cloud
        if job.backend.requirements.excluded_sites != []:
            taskParamMap['excludedSite'] = expandExcludedSiteList( job )

        # if only a single site specifed, don't set includedSite
        #if job.backend.site != 'AUTO':
        #    taskParamMap['includedSite'] = job.backend.site
        #taskParamMap['cliParams'] = fullExecString
        if job.backend.requirements.noEmail:
            taskParamMap['noEmail'] = True
        if job.backend.requirements.skipScout:
            taskParamMap['skipScout'] = True
        if not app.atlas_exetype in ["ATHENA", "TRF"]: 
            taskParamMap['nMaxFilesPerJob'] = job.backend.requirements.maxNFilesPerJob
        if job.backend.requirements.disableAutoRetry:
            taskParamMap['disableAutoRetry'] = 1
        # source URL
        matchURL = re.search("(http.*://[^/]+)/",Client.baseURLCSRVSSL)
        if matchURL is not None:
            taskParamMap['sourceURL'] = matchURL.group(1)

        # dataset names
        outDatasetName = job.outputdata.datasetname
        logDatasetName = re.sub('/$','.log/',job.outputdata.datasetname)
        # log
        taskParamMap['log'] = {'dataset': logDatasetName,
                               'container': logDatasetName,
                               'type':'template',
                               'param_type':'log',
                               'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName[:-1])
                               }
        # job parameters
        if app.atlas_exetype in ["ATHENA", "TRF"]:
            taskParamMap['jobParameters'] = [
                {'type':'constant',
                 'value': ' --sourceURL ${SURL}',
                 },
                ]
        else:
            taskParamMap['jobParameters'] = [
                {'type':'constant',
                 'value': '-j "" --sourceURL ${SURL}',
                 },
                ]

        taskParamMap['jobParameters'] += [
            {'type':'constant',
             'value': '-r {0}'.format(self.rundirectory),
             },
            ]

        # Add the --trf option to jobParameters if required
        if app.atlas_exetype == "TRF":
            taskParamMap['jobParameters'] += [{'type': 'constant', 'value': '--trf'}]

        # output
        # output files
        outMap = {}
        if app.atlas_exetype in ["ATHENA", "TRF"]:
            outMap, tmpParamList = AthenaUtils.convertConfToOutput(self.runConfig, self.extOutFile, job.outputdata.datasetname, destination=job.outputdata.location)
            taskParamMap['jobParameters'] += [
                {'type':'constant',
                 'value': '-o "%s" ' % outMap
                 },
                ]
            taskParamMap['jobParameters'] += tmpParamList 

        else: 
            if job.outputdata.outputdata:
                for tmpLFN in job.outputdata.outputdata:
                    if len(job.outputdata.datasetname.split('.')) > 2:
                        lfn = '{0}.{1}'.format(*job.outputdata.datasetname.split('.')[:2])
                    else:
                        lfn = job.outputdata.datasetname[:-1]
                    lfn += '.$JOBSETID._${{SN/P}}.{0}'.format(tmpLFN)
                    dataset = '{0}_{1}/'.format(job.outputdata.datasetname[:-1],tmpLFN)
                    taskParamMap['jobParameters'] += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True, destination=job.outputdata.location)
                    outMap[tmpLFN] = lfn

                taskParamMap['jobParameters'] += [ 
                    {'type':'constant',
                     'value': '-o "{0}"'.format(str(outMap)),
                     },
                    ]

        if app.atlas_exetype in ["ATHENA"]:
            # jobO parameter
            tmpJobO = self.job_options
            # replace full-path jobOs
            for tmpFullName,tmpLocalName in AthenaUtils.fullPathJobOs.iteritems():
                tmpJobO = re.sub(tmpFullName,tmpLocalName,tmpJobO)
            # modify one-liner for G4 random seeds
            if self.runConfig.other.G4RandomSeeds > 0:
                if app.options != '':
                    tmpJobO = re.sub('-c "%s" ' % app.options,
                                     '-c "%s;from G4AtlasApps.SimFlags import SimFlags;SimFlags.SeedsG4=${RNDMSEED}" ' \
                                         % app.options,tmpJobO)
                else:
                    tmpJobO = '-c "from G4AtlasApps.SimFlags import SimFlags;SimFlags.SeedsG4=${RNDMSEED}" '
                dictItem = {'type':'template',
                            'param_type':'number',
                            'value':'${RNDMSEED}',
                            'hidden':True,
                            'offset':self.runConfig.other.G4RandomSeeds,
                            }
                taskParamMap['jobParameters'] += [dictItem]
        elif app.atlas_exetype in ["TRF"]:
            # replace parameters for TRF
            tmpJobO = self.job_options
            # output : basenames are in outMap['IROOT'] trough extOutFile
            tmpOutMap = []
            for tmpName,tmpLFN in outMap['IROOT']:
                tmpJobO = tmpJobO.replace('%OUT.' + tmpName,tmpName)
            # replace DBR
            tmpJobO = re.sub('%DB=[^ \'\";]+','${DBR}',tmpJobO)

        if app.atlas_exetype in ["TRF"]:
            taskParamMap['useLocalIO'] = 1

        # build
        if job.backend.nobuild:
            taskParamMap['jobParameters'] += [
                {'type':'constant',
                 'value': '-a {0}'.format(os.path.basename(self.inputsandbox)),
                 },
                ]
        else:
            taskParamMap['jobParameters'] += [
                {'type':'constant',
                 'value': '-l ${LIB}',
                 },
                ]

        #
        # input
        if job.inputdata and job.inputdata._name == 'DQ2Dataset':
            if job.backend.requirements.nFilesPerJob > 0 and job.inputdata.number_of_files == 0 and job.backend.requirements.split > 0:
                job.inputdata.number_of_files = job.backend.requirements.nFilesPerJob * job.backend.requirements.split

        if job.inputdata and job.inputdata._name == 'DQ2Dataset' and job.inputdata.number_of_files != 0:
            taskParamMap['nFiles'] = job.inputdata.number_of_files
        elif job.backend.requirements.nFilesPerJob > 0 and job.backend.requirements.split > 0:
            # pathena does this for some reason even if there is no input files
            taskParamMap['nFiles'] = job.backend.requirements.nFilesPerJob * job.backend.requirements.split
        if job.backend.requirements.nFilesPerJob > 0:    
            taskParamMap['nFilesPerJob'] = job.backend.requirements.nFilesPerJob
            
        if job.backend.requirements.nEventsPerFile > 0:    
            taskParamMap['nEventsPerFile'] = job.backend.requirements.nEventsPerFile

        if not job.backend.requirements.nGBPerJob in [ 0,'MAX']:
            try:
                if job.backend.requirements.nGBPerJob != 'MAX':
                    job.backend.requirments.nGBPerJob = int(job.backend.requirements.nGBPerJob)
            except:
                logger.error("nGBPerJob must be an integer or MAX")
            # check negative                                                                                                                                                         
            if job.backend.requirements.nGBPerJob <= 0:
                logger.error("nGBPerJob must be positive")

            # don't set MAX since it is the defalt on the server side
            if not job.backend.requirements.nGBPerJob in [-1,'MAX']: 
                taskParamMap['nGBPerJob'] = job.backend.requirements.nGBPerJob

        if app.atlas_exetype in ["ATHENA", "TRF"]:
            inputMap = {}
            if job.inputdata and job.inputdata._name == 'DQ2Dataset':
                tmpDict = {'type':'template',
                           'param_type':'input',
                           'value':'-i "${IN/T}"',
                           'dataset': ','.join(job.inputdata.dataset),
                           'expand':True,
                           'exclude':'\.log\.tgz(\.\d+)*$',
                           }
                #if options.inputType != '':
                #    tmpDict['include'] = options.inputType
                taskParamMap['jobParameters'].append(tmpDict)
                taskParamMap['dsForIN'] = ','.join(job.inputdata.dataset)
                inputMap['IN'] = ','.join(job.inputdata.dataset)
            else:
                # no input
                taskParamMap['noInput'] = True
                if job.backend.requirements.split > 0:
                    taskParamMap['nEvents'] = job.backend.requirements.split
                else:
                    taskParamMap['nEvents'] = 1
                taskParamMap['nEventsPerJob'] = 1
                taskParamMap['jobParameters'] += [
                    {'type':'constant',
                     'value': '-i "[]"',
                     },
                    ]
        else:
            if job.inputdata and job.inputdata._name == 'DQ2Dataset':
                tmpDict = {'type':'template',
                           'param_type':'input',
                           'value':'-i "${IN/T}"',
                           'dataset': ','.join(job.inputdata.dataset),
                           'expand':True,
                           'exclude':'\.log\.tgz(\.\d+)*$',
                           }
               #if options.nSkipFiles != 0:
               #    tmpDict['offset'] = options.nSkipFiles
                taskParamMap['jobParameters'].append(tmpDict)
                taskParamMap['dsForIN'] = ','.join(job.inputdata.dataset)
            else:
                # no input
                taskParamMap['noInput'] = True
                if job.backend.requirements.split > 0:
                    taskParamMap['nEvents'] = job.backend.requirements.split
                else:
                    taskParamMap['nEvents'] = 1
                taskParamMap['nEventsPerJob'] = 1

        # param for DBR     
        if self.dbrelease != '':
            dbrDS = self.dbrelease.split(':')[0]
            # change LATEST to DBR_LATEST
            if dbrDS == 'LATEST':
                dbrDS = 'DBR_LATEST'
            dictItem = {'type':'template',
                        'param_type':'input',
                        'value':'--dbrFile=${DBR}',
                        'dataset':dbrDS,
                            }
            taskParamMap['jobParameters'] += [dictItem]
        # no expansion
        #if options.notExpandDBR:
        #dictItem = {'type':'constant',
        #            'value':'--noExpandDBR',
        #            }
        #taskParamMap['jobParameters'] += [dictItem]

        # secondary FIXME disabled
        self.secondaryDSs = {}
        if self.secondaryDSs != {}:
            inMap = {}
            streamNames = []
            for tmpDsName,tmpMap in self.secondaryDSs.iteritems():
                # make template item
                streamName = tmpMap['streamName']
                dictItem = MiscUtils.makeJediJobParam('${'+streamName+'}',tmpDsName,'input',hidden=True,
                                                      expand=True,include=tmpMap['pattern'],offset=tmpMap['nSkip'],
                                                      nFilesPerJob=tmpMap['nFiles'])
                taskParamMap['jobParameters'] += dictItem
                inMap[streamName] = 'tmp_'+streamName 
                streamNames.append(streamName)
            # make constant item
            strInMap = str(inMap)
            # set placeholders
            for streamName in streamNames:
                strInMap = strInMap.replace("'tmp_"+streamName+"'",'${'+streamName+'/T}')
            dictItem = {'type':'constant',
                        'value':'--inMap "%s"' % strInMap,
                        }
            taskParamMap['jobParameters'] += [dictItem]

        # misc
        jobParameters = ''
        # use Athena packages
        if app.atlas_exetype == 'ARES' or (app.atlas_exetype in ['PYARA','ROOT','EXE'] and app.useAthenaPackages):
            jobParameters += "--useAthenaPackages "
            
        # use RootCore
        if app.useRootCore or app.useRootCoreNoBuild:
            jobParameters += "--useRootCore "
            
        # use mana
        if app.useMana:
            jobParameters += "--useMana "
            if app.atlas_release != "":
                jobParameters += "--manaVer %s " % app.atlas_release
        # root
        if app.atlas_exetype in ['PYARA','ROOT','EXE'] and job.backend.requirements.rootver != '':
            rootver = re.sub('/','.', job.backend.requirements.rootver)
            jobParameters += "--rootVer %s " % rootver

        # write input to txt
        #if options.writeInputToTxt != '':
        #    jobParameters += "--writeInputToTxt %s " % options.writeInputToTxt
        # debug parameters
        #if options.queueData != '':
        #    jobParameters += "--overwriteQueuedata=%s " % options.queueData
        # JEM
        #if options.enableJEM:
        #    jobParameters += "--enable-jem "
        #    if options.configJEM != '':
        #        jobParameters += "--jem-config %s " % options.configJEM

        # set task param
        if jobParameters != '':
            taskParamMap['jobParameters'] += [ 
                {'type':'constant',
                 'value': jobParameters,
                 },
                ]

        # force stage-in
        if job.backend.accessmode == "LocalIO":
            taskParamMap['useLocalIO'] = 1

        # set jobO parameter
        if app.atlas_exetype in ["ATHENA", "TRF"]:
            taskParamMap['jobParameters'] += [
                {'type':'constant',
                 'value': '-j "',
                 'padding':False,
                 },
                ]
            taskParamMap['jobParameters'] += PsubUtils.convertParamStrToJediParam(tmpJobO,inputMap,job.outputdata.datasetname[:-1], True,False)
            taskParamMap['jobParameters'] += [
                {'type':'constant',
                 'value': '"',
                 },
                ]

        else:
            taskParamMap['jobParameters'] += [ {'type':'constant',
                                                'value': '-p "{0}"'.format(urllib.quote(self.job_options)),
                                                },
                                               ]

        # build step
        if not job.backend.nobuild:
            jobParameters = '-i ${IN} -o ${OUT} --sourceURL ${SURL} '

            if job.backend.bexec != '':
                jobParameters += ' --bexec "%s" ' % urllib.quote(job.backend.bexec)

            if app.atlas_exetype == 'ARES' or (app.atlas_exetype in ['PYARA','ROOT','EXE'] and app.useAthenaPackages):
                # use Athena packages
                jobParameters += "--useAthenaPackages "
            # use RootCore
            if app.useRootCore or app.useRootCoreNoBuild:
                jobParameters += "--useRootCore "

            # run directory
            if app.atlas_exetype in ['PYARA','ARES','ROOT','EXE']:
                jobParameters += '-r {0} '.format(self.rundirectory)
                
            # no compile
            #if options.noCompile:
            #    jobParameters += "--noCompile "
            # use mana
            if app.useMana:
                jobParameters += "--useMana "
                if app.atlas_release != "":
                    jobParameters += "--manaVer %s " % app.atlas_release

            # root
            if app.atlas_exetype in ['PYARA','ROOT','EXE'] and job.backend.requirements.rootver != '':
                rootver = re.sub('/','.', job.backend.requirements.rootver)
                jobParameters += "--rootVer %s " % rootver
                
            # cmt config
            if app.atlas_exetype in ['PYARA','ARES','ROOT','EXE']:
                if not app.atlas_cmtconfig in ['','NULL',None]:
                    jobParameters += " --cmtConfig %s " % app.atlas_cmtconfig
                                            
            
            #cmtConfig         = AthenaUtils.getCmtConfig(athenaVer=app.atlas_release, cmtConfig=app.atlas_cmtconfig)
            #if cmtConfig:
            #    jobParameters += "--cmtConfig %s " % cmtConfig
            # debug parameters
            #if options.queueData != '':
            #    jobParameters += "--overwriteQueuedata=%s " % options.queueData
            # set task param
            taskParamMap['buildSpec'] = {
                'prodSourceLabel':'panda',
                'archiveName':os.path.basename(self.inputsandbox),
                'jobParameters':jobParameters,
                }


        # enable merging
        if job.backend.requirements.enableMerge:
            jobParameters = '-r {0} '.format(self.rundirectory)
            if 'exec' in job.backend.requirements.configMerge and job.backend.requirements.configMerge['exec'] != '':
                jobParameters += '-j "{0}" '.format(job.backend.requirements.configMerge['exec'])
            if not job.backend.nobuild:
                jobParameters += '-l ${LIB} '
            else:
                jobParameters += '-a {0} '.format(os.path.basename(self.inputsandbox))
                jobParameters += "--sourceURL ${SURL} "
            jobParameters += '${TRN_OUTPUT:OUTPUT} ${TRN_LOG:LOG}'
            taskParamMap['mergeSpec'] = {}
            taskParamMap['mergeSpec']['useLocalIO'] = 1
            taskParamMap['mergeSpec']['jobParameters'] = jobParameters
            taskParamMap['mergeOutput'] = True    
            
        # Selected by Jedi
        #if not app.atlas_exetype in ['PYARA','ROOT','EXE']:
        #    taskParamMap['transPath'] = 'http://atlpan.web.cern.ch/atlpan/runAthena-00-00-12'

        logger.debug(taskParamMap)

        # upload sources
        if self.inputsandbox and not job.backend.libds:
            uploadSources(os.path.dirname(self.inputsandbox),os.path.basename(self.inputsandbox))

            if not self.inputsandbox == tmp_user_area_name:
                logger.info('Removing source tarball %s ...' % self.inputsandbox )
                os.remove(self.inputsandbox)

        return taskParamMap

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        '''prepare the subjob specific configuration'''

        job = app._getParent()

        logger.debug('AthenaJediRTHandler prepare called for %s', job.getFQID('.'))

#       in case of a simple job get the dataset content, otherwise subjobs are filled by the splitter
        
        return {}

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

for app in ['Athena', 'ExecutableDQ2', 'RootDQ2']:
    allHandlers.add(app,'Jedi',AthenaJediRTHandler)

from GangaCore.Utility.Config import getConfig, ConfigError
configDQ2 = getConfig('DQ2')
configJedi = getConfig('Jedi')

from GangaCore.Utility.logging import getLogger
logger = getLogger()



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PDF_Local.py,v 1.1 2009-05-10 16:42:03 karl Exp $
###############################################################################
# File: PDF_Local.py
# Author: K. Harrison
# Created: 070123

"""Module containing class dealing with preparation of jobs to run
   application for PDF analysis on Local backend"""

__author__  = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__    = "10 May 2009"
__version__ = "1.0"

from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.GPIDev.Lib.File import  File
from GangaCore.GPIDev.Lib.File import  FileBuffer
from GangaCore.Utility import logging
from GangaCore.Utility.files import fullpath

import os
import time

logger = logging.getLogger()
ptime = \
   "\"$(echo \"$(python -c \"import time; print '%.6f' % time.time()\")\")\""

class PDF_Local( IRuntimeHandler ):

   def prepare( self, app, appsubconfig, appmasterconfig, jobmasterconfig ):

      job = app.getJobObject()
      version = appsubconfig[ "version" ]
      outDir = appsubconfig[ "outDir" ]
      docDir = appsubconfig[ "docDir" ]
      elementList = docDir.split( os.sep )
      docList = appsubconfig[ "docList" ]
      softwareDir = appsubconfig[ "softwareDir" ]

      lineList = []
      inbox = []
      outbox = []

      headList, headBox = self.head( job = job, version = version, \
         softwareDir = softwareDir )
      lineList.extend( headList )
      outbox.extend( headBox )

      bodyList, bodyBox = self.body( job = job, docDir = docDir, \
         docList = docList )
      lineList.extend( bodyList )
      inbox.extend( bodyBox )

      tailList, tailBox = self.tail( job = job )
      lineList.extend( tailList )
      outbox.extend( tailBox )

      jobScript = "\n".join( lineList )
      jobWrapper = FileBuffer( "PDF.sh", jobScript, executable = 1 )

      outbox.extend( job.outputsandbox )

      return StandardJobConfig\
         ( exe = jobWrapper, inputbox = inbox, outputbox = outbox )

   def head( self, job = None, version = "", softwareDir = "" ):
      outbox = []
      lineList = \
         [
            "#!/bin/bash",
            "",
            "# Run script for PDF analysis",
            "# Created by Ganga - %s" % ( time.strftime( "%c" ) ),
            "",
            "JOB_START_TIME=\\",
            "%s" % ptime,
            "HOSTNAME=$(hostname -f)",
            "echo \"Job running on ${HOSTNAME}\"",
            "echo \"Processor architecture: $(arch)\"",
            "echo \"Start time: $(date)\"",
            "",
            "WORKDIR=$(pwd)",
            "export CAMTOLOGY_VERSION=\"%s\"" % version,
            "if [ -z ${VO_CAMONT_SW_DIR} ]; then",
            "   VO_CAMONT_SW_DIR=\"%s\"" % softwareDir,
            "fi",
         ]
      return ( lineList, outbox )

   def body( self, job = None, docDir = "", docList = [] ):

      lineList = []

      lineList.extend\
         ( [
         "DOWNLOAD_START_TIME=\\",
         "%s" % ptime,
         "N_DOWNLOAD_SUCCESS=0",
         "N_DOWNLOAD_FAILURE=0",
         ])

      xList, inbox, removeList  = self.bodyUrlList( docList = docList )   
      lineList.extend( xList )

      lineList.extend\
         ( [
         "DOWNLOAD_END_TIME=\\",
         "%s" % ptime,
         "DOWNLOAD_SIZE=$(du -b -s ${WORKDIR}/media | cut -f 1)",
         ])

      lineList.extend\
         ( [
         "cd ${CLASSIFIER_HOME}",
         "COMMAND=\"${VO_CAMONT_SW_DIR}/Camtoloty/version-${CAMTOLOGY_VERSION}"\
            +"/src/bin/pdf2fbSxml-inputDir.sh ${WORKDIR}/media\"",
         "COMMAND=\"./pdf2fbSxml-inputDir.sh ${WORKDIR}/media\"",
         "echo \"Performing PDF analysis with command:\"",
         "echo \"   ${COMMAND}\"",
         "EXECUTION_START_TIME=\\",
         "%s" % ptime,
         "${COMMAND}",
         "EXECUTION_END_TIME=\\",
         "%s" % ptime,
         ] )

      lineList.extend( removeList )

      return ( lineList, inbox )

   def bodyUrlList( self, docList = [] ):

      lineList = []
      removeList = []
      inbox = []

      for docUrl, relativePath in docList:

         if relativePath:
            docFile = os.path.basename( relativePath )
            docSubdir = os.path.join\
               ( "media", os.path.dirname( relativePath ) )
            docRemoteDir = os.path.join( "${WORKDIR}", docSubdir )
            docRemotePath = os.path.join( docRemoteDir, docFile )
            removeList.append( "rm %s" % docRemotePath )
            lineList.extend\
               ( [
               "mkdir -p %s" % docRemoteDir,
               "wget -q -P %s '%s'" % ( docRemoteDir, docUrl ),
               "if [ -f %s ]; then" % docRemotePath,
               "   N_DOWNLOAD_SUCCESS=$((${N_DOWNLOAD_SUCCESS}+1))",
               "else",
               "   N_DOWNLOAD_FAILURE=$((${N_DOWNLOAD_FAILURE}+1))",
               "fi",
#              "sleep 30s",
               ] )
         else:
            docFile = os.path.basename( docUrl )
            docSuffix = os.path.splitext( docFile )[ 1 ]  
            lineList.extend\
               ( [
               "wget -q %s -O %s" % ( docUrl, docFile ),
               "if [ -f %s ]; then" % docFile,
               "   DOC_MD5SUM=$(md5sum %s)" % docFile,
               "   DOC_MD5NAME=$(echo ${DOC_MD5SUM} | cut -d ' ' -f 1)",
               "   SUBDIR1=$(echo ${DOC_MD5NAME} | cut -c 1-3)",
               "   SUBDIR2=$(echo ${DOC_MD5NAME} | cut -c 4-6)",
               "   N_DOWNLOAD_SUCCESS=$((${N_DOWNLOAD_SUCCESS}+1))",
               ] )
            docSubdir = os.path.join( "media", "${SUBDIR1}", "${SUBDIR2}" )
            docRemoteDir = os.path.join( "${WORKDIR}", docSubdir )
            docRemotePath = os.path.join\
               ( docRemoteDir, "${DOC_MD5NAME}" + docSuffix )
            lineList.extend\
               ( [
               "   mkdir -p %s" % docRemoteDir,
               "   mv -f %s %s" % ( docFile, docRemotePath ),
               "else",
               "   N_DOWNLOAD_FAILURE=$((${N_DOWNLOAD_FAILURE}+1))",
               "fi",
#              "sleep 30s",
               ] )

      return ( lineList, inbox, removeList )

   def tail( self, job = None ):
      lineList = \
         [
         "",
         "RESULTS_SIZE=$(du -b -s ${WORKDIR}/media | cut -f 1)",
         "cd ${WORKDIR}/media",
         "TAR_START_TIME=\\",
         "%s" % ptime,
         "tar -zcf ${WORKDIR}/media.tar.gz *",
         "TAR_END_TIME=\\",
         "%s" % ptime,
         "TARBALL_SIZE=$(du -b -s ${WORKDIR}/media.tar.gz | cut -f 1)",
         "",
         "RUN_DATA=${WORKDIR}/execute.dat",
         "rm -rf ${RUN_DATA}",
         "touch ${RUN_DATA}",
         "echo \"Hostname: ${HOSTNAME}\" >> ${RUN_DATA}",
         "echo \"Job_start: ${JOB_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Download_start: ${DOWNLOAD_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Download_end: ${DOWNLOAD_END_TIME}\" >> ${RUN_DATA}",
         "echo \"Download_successes: ${N_DOWNLOAD_SUCCESS}\" >> ${RUN_DATA}",
         "echo \"Download_failures: ${N_DOWNLOAD_FAILURE}\" >> ${RUN_DATA}",
         "echo \"Download_size: ${DOWNLOAD_SIZE}\" >> ${RUN_DATA}",
         "echo \"Execution_start: ${EXECUTION_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Execution_end: ${EXECUTION_END_TIME}\" >> ${RUN_DATA}",
         "echo \"Results_size: ${RESULTS_SIZE}\" >> ${RUN_DATA}",
         "echo \"Tar_start: ${TAR_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Tar_end: ${TAR_END_TIME}\" >> ${RUN_DATA}",
         "echo \"Tarball_size: ${TARBALL_SIZE}\" >> ${RUN_DATA}",
         "JOB_END_TIME=\\",
         "%s" % ptime,
         "echo \"Job_end: ${JOB_END_TIME}\" >> ${RUN_DATA}",
         "echo \"End time: $(date)\"",
         ]
      outbox = [ "media.tar.gz", "execute.dat" ]
      return ( lineList, outbox )
